import sys
import qdarkstyle
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import (
    QMainWindow, QApplication, QWidget, QVBoxLayout, QTabWidget, QStatusBar, QToolBar,
    QAction, QHBoxLayout, QGroupBox, QCheckBox
)

from marge.controller.controller_console import ConsoleController
from marge.widgets.widget_hw_console import ConsoleWidget
from marge.widgets.widget_hw_others import OthersWidget
from marge.widgets.widget_hw_gradients import GradientsWidget
from marge.widgets.widget_hw_rf import RfWidget
from marge.widgets.widget_session import SessionWidget


class SessionWindow(QMainWindow):
    def __init__(self):
        super().__init__()

        self.setWindowTitle("Session and Hardware Window for MaRGE")
        self.setGeometry(100, 100, 600, 400)

        # Set initial stylesheet
        self.is_dark_theme = True
        self.setStyleSheet(qdarkstyle.load_stylesheet_pyqt5())

        # Main layout
        widget_main = QWidget()
        layout_main = QHBoxLayout()
        widget_main.setLayout(layout_main)
        self.setCentralWidget(widget_main)
        layout_left = QVBoxLayout()
        layout_main.addLayout(layout_left)

        # Scanner checks
        checks = QGroupBox("Scanner checks")
        layout = QVBoxLayout()
        checks.setLayout(layout)
        layout_left.addWidget(checks)

        self.check_projects = QCheckBox("Projects")
        self.check_projects.setDisabled(True)
        layout.addWidget(self.check_projects)

        self.check_study = QCheckBox("Study cases")
        self.check_study.setDisabled(True)
        layout.addWidget(self.check_study)

        self.check_gradients = QCheckBox("Gradient hardware")
        self.check_gradients.setDisabled(True)
        layout.addWidget(self.check_gradients)

        self.check_redpitayas = QCheckBox("MaRCoS")
        self.check_redpitayas.setDisabled(True)
        layout.addWidget(self.check_redpitayas)

        self.check_rf = QCheckBox("RFs")
        self.check_rf.setDisabled(True)
        layout.addWidget(self.check_rf)

        self.check_rp_ips = QCheckBox("RP IPs")
        self.check_rp_ips.setDisabled(True)
        layout.addWidget(self.check_rp_ips)

        self.check_rf_coils = QCheckBox("RF coils")
        self.check_rf_coils.setDisabled(True)
        layout.addWidget(self.check_rf_coils)

        self.check_others = QCheckBox("Others")
        self.check_others.setDisabled(True)
        layout.addWidget(self.check_others)

        # Console
        self.console = ConsoleController()
        layout_left.addWidget(self.console)

        # Toolbar
        self.toolbar = QToolBar("Session toolbar")
        self.addToolBar(self.toolbar)

        self.launch_gui_action = QAction(QIcon("resources/icons/home.png"), "Launch GUI", self)
        self.launch_gui_action.setStatusTip("Launch GUI")
        self.launch_gui_action.setDisabled(True)
        self.toolbar.addAction(self.launch_gui_action)

        self.demo_gui_action = QAction(QIcon("resources/icons/demo.png"), "Launch GUI as DEMO", self)
        self.demo_gui_action.setStatusTip("Launch GUI as DEMO")
        self.demo_gui_action.setDisabled(True)
        self.toolbar.addAction(self.demo_gui_action)

        self.update_action = QAction(QIcon("resources/icons/arrow-sync.svg"), "Update scanner hardware", self)
        self.update_action.setStatusTip("Update scanner hardware")
        self.toolbar.addAction(self.update_action)

        self.close_action = QAction(QIcon("resources/icons/close.png"), "Close GUI", self)
        self.close_action.setStatusTip("Close the GUI")
        self.toolbar.addAction(self.close_action)

        # Add switch theme button
        self.switch_theme_action = QAction(QIcon("resources/icons/adjust-contrast.svg"), "", self)
        self.switch_theme_action.setStatusTip("Switch between Dark and Light theme")
        self.toolbar.addAction(self.switch_theme_action)

        # Tabs
        self.tabs = QTabWidget()
        layout_main.addWidget(self.tabs)

        self.tab_session = SessionWidget()
        self.tab_console = ConsoleWidget()
        self.tab_gradients = GradientsWidget()
        self.tab_rf = RfWidget()
        self.tab_others = OthersWidget()

        self.tabs.addTab(self.tab_session, "Session")
        self.tabs.addTab(self.tab_console, "Console")
        self.tabs.addTab(self.tab_gradients, "Gradients")
        self.tabs.addTab(self.tab_rf, "RF")
        self.tabs.addTab(self.tab_others, "Others")

        # Status bar
        self.setStatusBar(QStatusBar(self))

    def setupTab1(self):
        widget = ConsoleWidget()
        layout = QVBoxLayout()
        layout.addWidget(widget)
        self.tab1.setLayout(layout)

    def setupTab2(self):
        widget = GradientsWidget()
        layout = QVBoxLayout()
        layout.addWidget(widget)
        self.tab2.setLayout(layout)

    def setupTab3(self):
        widget = RfWidget()
        layout = QVBoxLayout()
        layout.addWidget(widget)
        self.tab3.setLayout(layout)

    def setupTab4(self):
        widget = OthersWidget()
        layout = QVBoxLayout()
        layout.addWidget(widget)
        self.tab4.setLayout(layout)


if __name__ == '__main__':
    app = QApplication(sys.argv)
    window = SessionWindow()
    window.show()
    app.exec()



#  NanoVNASaver
#
#  A python program to view and export Touchstone data from a NanoVNA
#  Copyright (C) 2020,2021 NanoVNA-Saver Authors
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
import logging

from marge.vna.Serial import Interface
from marge.vna.NanoVNA_H import NanoVNA_H

logger = logging.getLogger(__name__)


class NanoVNA_H4(NanoVNA_H):
    name = "NanoVNA-H4"
    screenwidth = 480
    screenheight = 320
    valid_datapoints = (101, 11, 51, 201, 401)

    def __init__(self, iface: Interface):
        super().__init__(iface)
        self.sweep_max_freq_Hz = 1500e6
        self.sweep_method = "scan"
        if "Scan mask command" in self.features:
            self.sweep_method = "scan_mask"

    # def read_features(self):
    #     logger.debug("read_features")
    #     super().read_features()
    #     if self.readFirmware().find("DiSlord") > 0:
    #         self.features.add("Customizable data points")
    #         logger.info("VNA has 201 datapoints capability")
    #         self.valid_datapoints = (201, 11, 51,101)



from PyQt5.QtWidgets import QTabWidget

from marge.controller.controller_post import PostProcessingTabController
from controller.controller_preprocessing import PreProcessingTabController
from marge.controller.controller_reconstruction import ReconstructionTabController
from marge.controller.controller_visualisation import VisualisationTabController


class ProcessingWidget(QTabWidget):
    """
    TabWidget class for displaying a tab widget with different tabs.

    Inherits from QTabWidget provided by PyQt5 to display a tab widget with different tabs.

    Attributes:
        main: The parent widget.
    """

    def __init__(self, parent, *args, **kwargs):
        """
        Initialize the TabWidget.

        Args:
            parent: The parent widget.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        super(ProcessingWidget, self).__init__(*args, **kwargs)
        self.main = parent

        self.setMaximumWidth(400)

        # Tabs
        self.main.preprocessing_controller = PreProcessingTabController(self.main)
        self.main.reconstruction_controller = ReconstructionTabController(self.main)
        self.main.postprocessing_controller = PostProcessingTabController(self.main)
        self.main.visualisation_controller = VisualisationTabController(self.main)

        # Adding Tabs in the QTabWidget
        self.addTab(self.main.preprocessing_controller, 'Pre-Processing')
        self.addTab(self.main.reconstruction_controller, 'Reconstruction')
        self.addTab(self.main.postprocessing_controller, 'Post-Processing')
        self.addTab(self.main.visualisation_controller, 'Visualisation')



from pyqtgraph import ImageView


class ImageViewWidget(ImageView):
    """
    ImageViewWidget class for displaying an image view.

    Inherits from ImageView provided by pyqtgraph to display an image view.

    Attributes:
        main: The parent widget.
    """

    def __init__(self, parent, *args, **kwargs):
        """
        Initialize the ImageViewWidget.

        Args:
            parent: The parent widget.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        super(ImageViewWidget, self).__init__(*args, **kwargs)
        self.main = parent

        self.setMinimumSize(400, 400)



#!/usr/bin/env python3
#
# Basic toolbox for server operations; wraps up a lot of stuff to avoid the need for hardcoding on the user's side.

import socket, time, warnings
import numpy as np
import matplotlib.pyplot as plt

from local_config import ip_address, port, fpga_clk_freq_MHz, grad_board
import grad_board as gb
import server_comms as sc
import marcompile as fc

import pdb
st = pdb.set_trace

######## TODO: configure the final buffers as well, whether in marcompile or elsewhere

class Experiment:
    """Wrapper class for managing an entire experimental sequence

    lo_freq: local oscillator frequencies, MHz: either single float,
    iterable of two or iterable of three values. Control three
    independent LOs. At least a single float must be supplied.
    If supplying 2 or 3 values, must also specify the rx_lo.

    rx_t: RF RX sampling time/s in microseconds; single float or tuple
    of two.

    rx_lo: RX local oscillator sources (integers): 0 - 2 correspond to
    the three LO NCOs, 3 is DC. Single integer or tuple of two.  By
    default will use local oscillator 0 for both RX channels, unless
    otherwise specified.

    Note: TX0 and TX1 will always be assigned to NCOs 0 and 1. NCO 2
    is for free adjustment of the RX frequency relative to the TX.

    csv_file: path to a CSV execution file, which will be
    compiled with marcompile.py . If this is not supplied, the
    sequence bytecode should be supplied manually using
    define_sequence() before run() is called.

    OPTIONAL PARAMETERS - ONLY ALTER IF YOU KNOW WHAT YOU'RE DOING

    grad_max_update_rate: used to calculate the frequency to run the
    gradient SPI interface at - must be high enough to support your
    maximum gradient sample rate but not so high that you experience
    communication issues. Leave this alone unless you know what you're
    doing.

    print_infos: print debugging messages from server to stdout

    assert_errors: errors returned from the server will be treated as
    exceptions by the class, halting the program

    init_gpa: initialise the GPA during the construction of this class

    """

    def __init__(self,
                 lo_freq=1, # MHz
                 rx_t=3.125, # us; multiples of 1/122.88, such as 3.125, are exact, others will be rounded to the nearest multiple of the 122.88 MHz clock
                 seq_dict=None,
                 seq_csv=None,
                 rx_lo=0, # which of internal NCO local oscillators (LOs), out of 0, 1, 2, to use for each channel
                 grad_max_update_rate=0.2, # MSPS, across all channels in parallel, best-effort
                 gpa_fhdo_offset_time=0, # when GPA-FHDO is used, offset the Y, Z and Z2 gradient times by 1x, 2x and 3x this value to emulate 'simultaneous' updates
                 print_infos=True, # show server info messages
                 assert_errors=True, # halt on server errors
                 init_gpa=False, # initialise the GPA (will reset its outputs when the Experiment object is created)
                 initial_wait=None, # initial pause before experiment begins - required to configure the LOs and RX rate; must be at least a few us. Is suitably set based on grad_max_update_rate by default.
                 auto_leds=True, # automatically scan the LED pattern from 0 to 255 as the sequence runs (set to off if you wish to manually control the LEDs)
                 prev_socket=None, # previously-opened socket, if want to maintain status etc
                 fix_cic_scale=True, # scale the RX data precisely based on the rate being used; otherwise a 2x variation possible in data amplitude based on rate
                 set_cic_shift=False, # program the CIC internal bit shift to maintain the gain within a factor of 2 independent of rate; required if the open-source CIC is used in the design
                 allow_user_init_cfg=False, # allow user-defined alteration of marga configuration set by init, namely RX rate, LO properties etc; see the compile() method for details
                 halt_and_reset=False, # upon connecting to the server, halt any existing sequences that may be running
                 flush_old_rx=False, # when debugging or developing new code, you may accidentally fill up the RX FIFOs - they will not automatically be cleared in case there is important data inside. Setting this true will always read them out and clear them before running a sequence. More advanced manual code can read RX from existing sequences.
                 ):

        # create socket early so that destructor works
        self._close_socket = True
        if prev_socket is None:
            self._s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self._s.connect( (ip_address, port) )
        else:
            self._s = prev_socket
            self._close_socket = False # do not close previous socket

        self.set_lo_freq(lo_freq)

        if not hasattr(rx_t, "__len__"):
            rx_t = rx_t, rx_t # extend to 2 elements

        # TODO: enable variable rates during a single TR
        self._rx_divs = np.round(np.array(rx_t) * fpga_clk_freq_MHz).astype(np.uint32)
        self._rx_ts = self._rx_divs / fpga_clk_freq_MHz

        if not hasattr(rx_lo, "__len__"):
            rx_lo = rx_lo, rx_lo # extend to 2 elements
        self._rx_lo = rx_lo

        assert grad_board in ('ocra1', 'gpa-fhdo'), "Unknown gradient board!"
        if grad_board == 'ocra1':
            gradb_class = gb.OCRA1
            self._gpa_fhdo_offset_time = 0
        else:
            gradb_class = gb.GPAFHDO
            self._gpa_fhdo_offset_time = gpa_fhdo_offset_time
        self.gradb = gradb_class(self.server_command, grad_max_update_rate)

        if initial_wait is None:
            # auto-set the initial wait to be long enough for initial gradient configuration to finish, plus 1us for miscellaneous startup
            self._initial_wait = 1 + 1/grad_max_update_rate

        self._auto_leds = auto_leds

        assert (seq_csv is None) or (seq_dict is None), "Cannot supply both a sequence dictionary and a CSV file."
        self._csv = None
        self._seq = None
        if seq_dict is not None:
            self.add_flodict(seq_dict)
        elif seq_csv is not None:
            self._csv = seq_csv # None unless a CSV was supplied

        self._print_infos = print_infos
        self._assert_errors = assert_errors

        if init_gpa:
            self.gradb.init_hw()

        if halt_and_reset:
            halted = sc.command({'halt_and_reset': 0}, self._s)[0][4]['halt_and_reset']
            assert halted, "Could not halt the execution of an existing sequence. Please file a bug report."

        self._fix_cic_scale = fix_cic_scale
        self._set_cic_shift = set_cic_shift
        self._flush_old_rx = flush_old_rx
        self._allow_user_init_cfg = allow_user_init_cfg

    def __del__(self):
        if self._close_socket:
            self._s.close()

    def server_command(self, server_dict):
        return sc.command(server_dict, self._s, self._print_infos, self._assert_errors)

    def get_rx_ts(self):
        return self._rx_ts

    def set_lo_freq(self, lo_freq):
        # lo_freq: either a single floating-point value, or an iterable of up to three values for each marga NCO

        # extend lo_freq to 3 elements
        if not hasattr(lo_freq, "__len__"):
            lo_freq = lo_freq, lo_freq, lo_freq # extend to 3 elements
        elif len(lo_freq) < 3:
            lo_freq = lo_freq[0], lo_freq[1], lo_freq[0] # extend from 2 to 3 elements

        self._dds_phase_steps = np.round(2**31 / fpga_clk_freq_MHz * np.array(lo_freq)).astype(np.uint32)
        self._lo_freqs = self._dds_phase_steps * fpga_clk_freq_MHz / (2 ** 31) # real LO freqs -- TODO: print for debugging

        self._seq_compiled = False # force recompilation

    def flo2int(self, seq_dict):
        """Convert a floating-point sequence dictionary to an integer binary
        dictionary"""

        intdict = {}

        ## Various functions to handle the conversion
        def times_us(farr):
            """ farr: float array, times in us units; [0, inf) """
            return np.round(fpga_clk_freq_MHz * farr).astype(np.int64) # negative values will get rejected at a later stage

        def tx_real(farr):
            """ farr: float array, [-1, 1] """
            return np.round(32767 * farr).astype(np.uint16)

        def tx_complex(times, farr, tolerance=2e-6):
            """times: float time array, farr: complex float array, [-1-1j, 1+1j]
            tolerance: minimum difference two values need to be considered binary-unique (2e-6 corresponds to ~19 bits)
            -- returns a tuple with repeated elements removed"""
            idata, qdata = farr.real, farr.imag
            unique = lambda k: np.concatenate([[True], np.abs(np.diff(k)) > tolerance])
            # DEBUGGING: use the below lambda instead to avoid stripping repeated values
            # unique = lambda k: np.ones_like(k, dtype=bool)
            idata_u, qdata_u = unique(idata), unique(qdata)
            tbins = ( times_us(times[idata_u] + self._initial_wait), times_us(times[qdata_u] + self._initial_wait) )
            txbins = ( tx_real(idata[idata_u]), tx_real(qdata[qdata_u]) )
            return tbins, txbins

        for key, (times, vals) in seq_dict.items():
            # each possible dictionary entry returns a tuple (even if one element) for the binary dictionary to send to marcompile
            tbin = times_us(times + self._initial_wait),
            if key in ['tx0_i', 'tx0_q', 'tx1_i', 'tx1_q']:
                valbin = tx_real(vals),
                keybin = key,
            elif key in ['tx0', 'tx1']:
                tbin, valbin = tx_complex(times, vals)
                keybin = key + '_i', key + '_q'
            elif key in ['grad_vx', 'grad_vy', 'grad_vz', 'grad_vz2',
                         'fhdo_vx', 'fhdo_vy', 'fhdo_vz', 'fhdo_vz2',
                         'ocra1_vx', 'ocra1_vy', 'ocra1_vz', 'ocra1_vz2']:
                # marcompile will figure out whether the key matches the selected grad board
                keyb, channel = self.gradb.key_convert(key)
                keybin = keyb, # tuple
                valbin = self.gradb.float2bin(vals, channel),

                # hack to de-synchronise GPA-FHDO outputs, to give the
                # user the illusion of being able to output on several
                # channels in parallel
                if self._gpa_fhdo_offset_time:
                    tbin = times_us(times + channel*self._gpa_fhdo_offset_time + self._initial_wait),

            elif key in ['rx0_rate', 'rx1_rate']:
                keybin = key,
                valbin = vals.astype(np.uint16),
            elif key in ['rx0_rate_valid', 'rx1_rate_valid', 'rx0_rst_n', 'rx1_rst_n', 'rx0_en', 'rx1_en', 'tx_gate', 'rx_gate', 'trig_out']:
                keybin = key,
                # binary-valued data
                valbin = vals.astype(np.int32),
                for vb in valbin:
                    assert np.all( (0 <= vb) & (vb <= 1) ), "Binary columns must be [0,1] or [False, True] valued"
            elif key in ['leds']:
                keybin = key,
                valbin = vals.astype(np.int32),
            else:
                warnings.warn("Unknown marga experiment dictionary key: " + key)
                continue

            for t, k, v in zip(tbin, keybin, valbin):
                intdict[k] = (t, v)

        return intdict

    def add_intdict(self, seq_intdict, append=True):
        """ Add an integer-format dictionary to the sequence, or replace the old (time, value) tuples with new ones """
        if self._seq is None:
            self._seq = {}

        for name, sb in seq_intdict.items():
            if name in self._seq.keys() and append:
                a, b = self._seq[name]
                self._seq[name] = ( np.append(a, sb[0]), np.append(b, sb[1]) )
            else:
                self._seq[name] = sb

    def add_flodict(self, flodict, append=True):
        """ Add a floating-point dictionary to the sequence """
        assert self._csv is None, "Cannot replace the dictionary for an Experiment class created from a CSV"
        self.add_intdict(self.flo2int(flodict), append)
        self._seq_compiled = False

    def compile(self):
        """Convert either dictionary or CSV file into machine code, with
        extra machine code at the start to ensure the system is initialised to
        the correct state.

        Initially, configure the RX rates and set the LEDs.
        Remainder of the sequence will be as programmed.
        """

        # RX and LO configuration
        tstart = 50 # cycles before doing anything
        rx_wait = 50 # cycles to run RX before setting rate, then later resetting again
        initial_cfg = {# 'rx0_rst_n': ( np.array([tstart, tstart + 2*rx_wait]), np.array([1, 0]) ),
                       # 'rx1_rst_n': ( np.array([tstart, tstart + 2*rx_wait]), np.array([1, 0]) ),
                       'rx0_rst_n': ( np.array([tstart]), np.array([1]) ),
                       'rx1_rst_n': ( np.array([tstart]), np.array([1]) ),
                       'lo0_freq': ( np.array([tstart]), np.array([self._dds_phase_steps[0]]) ),
                       'lo1_freq': ( np.array([tstart]), np.array([self._dds_phase_steps[1]]) ),
                       'lo2_freq': ( np.array([tstart]), np.array([self._dds_phase_steps[2]]) ),
                       'lo0_rst': ( np.array([tstart, tstart + 1]), np.array([1, 0]) ),
                       'lo1_rst': ( np.array([tstart, tstart + 1]), np.array([1, 0]) ),
                       'lo2_rst': ( np.array([tstart, tstart + 1]), np.array([1, 0]) )
                       }

        # Set CIC decimation rate and internal shift, if necessary, and calculate CIC scale correction
        rx0_words, self._rx0_cic_factor = fc.cic_words(self._rx_divs[0], self._set_cic_shift)
        rx1_words, self._rx1_cic_factor = fc.cic_words(self._rx_divs[1], self._set_cic_shift)
        if not self._fix_cic_scale: # clear the correction factor
            self._rx0_cic_factor = 1
            self._rx1_cic_factor = 1
        rx0r_st = tstart + rx_wait
        wds = len(rx0_words)
        ar0 = np.arange(wds, dtype=int)
        ar1 = np.arange(wds + 1, dtype=int)
        ow = np.ones(wds + 1, dtype=int)
        ow[-1] = 0

        initial_cfg.update({
            'rx0_rate': ( tstart + rx_wait + ar0, np.array(rx0_words) ),
            'rx1_rate': ( tstart + rx_wait + ar0, np.array(rx1_words) ),
            'rx0_rate_valid': ( tstart + rx_wait + ar1, ow),
            'rx1_rate_valid': ( tstart + rx_wait + ar1, ow),
        })

        # LO source configuration (only set non-default values if necessary)
        if self._rx_lo[0] != 0:
            initial_cfg.update({ 'rx0_lo': ( np.array([tstart]), np.array([self._rx_lo[0]]) ) })
        if self._rx_lo[1] != 0:
            initial_cfg.update({ 'rx1_lo': ( np.array([tstart]), np.array([self._rx_lo[1]]) ) })

        # Automatic LED scan
        if self._auto_leds:
            led_steps = 256

            # find max time used in system
            ultimate_time = 0
            for k in self._seq:
                loc_last_time = self._seq[k][0][-1]
                if loc_last_time > ultimate_time:
                    ultimate_time = loc_last_time

            if ultimate_time < 256:
                led_steps = ultimate_time
            led_times = np.linspace(tstart, ultimate_time + tstart, 256).astype(np.int64)
            led_vals = np.linspace(1, 256, led_steps).astype(np.uint32)
            initial_cfg.update({ 'leds': (led_times, led_vals) }) # should overlap with any previous LED settings

        # do not clear relevant dictionary values if user-defined configuration of init parameters at runtime is allowed
        self.add_intdict(initial_cfg, append=self._allow_user_init_cfg)

        self._machine_code = np.array( fc.dict2bin(self._seq,
                                             self.gradb.bin_config['initial_bufs'],
                                             self.gradb.bin_config['latencies'], # TODO: can add extra manipulation here, e.g. add to another array etc
                                             ), dtype=np.uint32 )

        self._seq_compiled = True

    def get_flodict(self, intd=None):
        """Calculate floating-point dictionaries based on the data inside the
        Experiment class so far -- useful for plotting or testing the sequence"""

        if intd is None:
            if not self._seq_compiled:
                self.compile()

            intd = self._seq

        flodict = {}

        def convert_t(t_bin, y):
            # add a zero event in the beginning, and shift the times to the 'user frame'
            t = np.concatenate( ([0], t_bin) ) /fpga_clk_freq_MHz - self._initial_wait
            # add a zero value in the beginning of outputs
            y2 = np.concatenate( ([0], y) )
            return t, y2

        # Convert TX channels
        for txl in ['tx0_i', 'tx0_q', 'tx1_i', 'tx1_q']:
            try:
                t_bin, tx_bin = intd[txl]
                t, tx = convert_t(t_bin, tx_bin.astype(np.int16) / 32768)
                flodict[txl] = (t, tx)
            except KeyError:
                continue

        # Convert gradient channels
        for gradl in self.gradb.keys():
            try:
                t_bin, grad_bin = intd[gradl]
                t, grad = convert_t(t_bin, self.gradb.bin2float(grad_bin) )
                flodict[gradl] = (t, grad)
            except KeyError:
                continue

        # Convert RX enable channels
        for rxl in ['rx0_en', 'rx1_en']:
            try:
                t_bin, rx = intd[rxl]
                t, rx = convert_t(t_bin, rx)
                flodict[rxl] = (t, rx)
            except KeyError:
                continue

        # Convert digital outputs
        for iol in ['tx_gate', 'rx_gate', 'trig_out', 'leds']:
            try:
                t_bin, io = intd[iol]
                t, io = convert_t(t_bin, io)
                if iol == 'leds':
                    io = io.astype(np.uint8).astype(float) / 256
                    flodict[iol] = (t, io)
            except KeyError:
                continue

        return flodict

    def plot_sequence(self, axes=None):
        """ axes: 4-element tuple of axes upon which the TX, gradients, RX and digital I/O plots will be drawn.
        If not provided, plot_sequence() will create its own. """
        if axes is None:
            _, axes = plt.subplots(4, 1, figsize=(12,8), sharex='col')

        (txs, grads, rxs, ios) = axes

        fd = self.get_flodict()

        # Plot TX channels
        for txl in ['tx0_i', 'tx0_q', 'tx1_i', 'tx1_q']:
            try:
                txs.step(*fd[txl], where='post', label=txl)
            except KeyError:
                continue

        # Plot gradient channels
        for gradl in self.gradb.keys():
            try:
                grads.step(*fd[gradl], where='post', label=gradl)
            except KeyError:
                continue

        # Plot RX enable channels
        for rxl in ['rx0_en', 'rx1_en']:
            try:
                rxs.step(*fd[rxl], where='post', label=rxl)
            except KeyError:
                continue

        # Plot digital outputs
        for iol in ['tx_gate', 'rx_gate', 'trig_out', 'leds']:
            try:
                ios.step(*fd[iol], where='post', label=iol)
            except KeyError:
                continue

        for ax in axes:
            ax.legend()
            ax.grid(True)

        ios.set_xlabel(r'time ($\mu$s)')
        return fd

    def run(self):
        """ compile the TX and grad data, send everything over.
        Returns the resultant data """

        if not self._seq_compiled:
            self.compile()

        if self._flush_old_rx:
            rx_data_old, _ = sc.command({'read_rx': 0}, self._s)
            # TODO: do something with RX data previously collected by the server

        rx_data, msgs = sc.command({'run_seq': self._machine_code.tobytes()}, self._s)

        rxd = rx_data[4]['run_seq']
        rxd_iq = {}

        # (1 << 24) just for the int->float conversion to be reasonable - exact value doesn't matter for now
        rx0_norm_factor = self._rx0_cic_factor / (1 << 24)
        rx1_norm_factor = self._rx0_cic_factor / (1 << 24)

        try:
            rxd_iq['rx0'] = rx0_norm_factor * ( np.array(rxd['rx0_i']).astype(np.int32).astype(float) + \
                             1j * np.array(rxd['rx0_q']).astype(np.int32).astype(float) )
        except (KeyError, TypeError):
            pass

        try:
            rxd_iq['rx1'] = rx1_norm_factor * ( np.array(rxd['rx1_i']).astype(np.int32).astype(float) + \
                             1j * np.array(rxd['rx1_q']).astype(np.int32).astype(float) )
        except (KeyError, TypeError):
            pass

        return rxd_iq, msgs

    def close_server(self, only_if_sim=False):
        ## Either always close server, or only close server if it's a simulation
        if not only_if_sim or sc.command({'are_you_real':0}, self._s)[0][4]['are_you_real'] == "simulation":
            sc.send_packet(sc.construct_packet({}, 0, command=sc.close_server_pkt), self._s)

def test_rx_scaling(lo_freq=0.5, rf_amp=0.5, rf_steps=True, rx_time=50, rx_periods=[600], rx_padding=20, plot_rx=False):

    expt = Experiment(lo_freq=lo_freq, rx_t=rx_periods[0] / fpga_clk_freq_MHz,
                      fix_cic_scale=False, set_cic_shift=False, allow_user_init_cfg=True, flush_old_rx=True)
    tr_t = 0
    tr_period = rx_time + rx_padding
    rx_lengths = []

    def single_pulse_tr(tstart, rx_period=20):
        if rf_steps:
            tx_shape = np.array([0, 0.2, 0.4, 0.6, 0.8, 1, 0.8, 0.6, 0.4, 0.2, 0]) * rf_amp
            tx_times = np.array([0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1]) * (rx_time - 2*rx_padding)
        else:
            tx_shape = np.array([1, 0]) * rf_amp
            tx_times = np.array([0, 1]) * (rx_time - 2*rx_padding)
        tx_seq = ( tstart + 5 + rx_padding + tx_times, tx_shape )
        rx_seq = ( tstart + 5 + np.array([0, rx_time]), np.array([1, 0]) )

        # Rate adjustment
        set_cic_shift = expt._set_cic_shift
        rx_words, _ = fc.cic_words(rx_period, set_cic_shift)
        rx_wait = 100
        rxr_st = tstart + rx_wait
        wds = len(rx_words)
        ar0 = np.arange(wds, dtype=int)
        ar1 = np.arange(wds + 1, dtype=int)
        ow = np.ones(wds + 1, dtype=int)
        ow[-1] = 0
        rate_seq = ( tstart + (rx_wait + ar0) / fpga_clk_freq_MHz, np.array(rx_words) )
        rate_en_seq = ( tstart + (rx_wait + ar1) / fpga_clk_freq_MHz, ow )

        value_dict = {
            'tx0': tx_seq, 'tx1': tx_seq,
            'rx0_en': rx_seq,
            'rx1_en': rx_seq,
            'rx0_rate': rate_seq, 'rx1_rate': rate_seq,
            'rx0_rate_valid': rate_en_seq, 'rx1_rate_valid': rate_en_seq,
            }

        # if rx_period % 2:
        #     value_dict.pop('rx0_rate_valid', None)
        #     value_dict.pop('rx1_rate_valid', None)

        return value_dict

    for rt in rx_periods:
        expt.add_flodict( single_pulse_tr( tr_t , rt) )
        tr_t += tr_period
        rx_lengths.append( int(rx_time * fpga_clk_freq_MHz / rt) + 1)

    rxd, msgs = expt.run()
    expt.close_server(True)

    # rx_lengths_a = np.array(rx_lengths)
    # print("RX lengths from calculation: ", rx_lengths)

    pulse_means = []

    if plot_rx:
        rx0_mag = np.abs(rxd['rx0'])
        rx1_mag = np.abs(rxd['rx1'])

        initial_excess = 5 # amount to throw away from the start of each acquisition
        M = 0

        plt.figure(figsize=(12, 6))
        plt.xlabel('RX sample')
        plt.ylabel('RF magnitude')
        for k, l in zip(rx_periods, rx_lengths):
            x = np.arange(M + initial_excess, M+l)
            y = np.array(rx1_mag[M + initial_excess : M+l])

            _, cic_correction = fc.cic_words(k)
            ysc = y * cic_correction

            # avoid off-by-1 errors - bit of a hack
            if len(ysc) != len(x):
                x = x[:-1]

            if True:
                plt.plot(x, ysc)
            M += l
            # pulse_means.append( y[y > 0.5*y.max()].mean() )
            # pulse_means.append( y.mean() )
            pulse_means.append( ysc.max() )

        plt.figure(figsize=(8,7))
        pma = np.array(pulse_means)
        plt.subplot(2,1,1)
        plt.plot(rx_periods, pma/pma.mean())
        plt.xlabel('RX decimation')
        plt.ylabel('RX peak amp')
        plt.grid(True)
        # plt.subplot(3,1,2)
        # plt.plot(rx_periods, rx_lengths)
        # plt.legend(['rx samples'])
        # plt.xlabel('RX decimation')
        # plt.grid(True)
        plt.subplot(2,1,2)
        _, cic_corrections = fc.cic_words(rx_periods)
        plt.plot(rx_periods, cic_corrections )
        plt.xlabel('RX decimation')
        plt.ylabel('RX CIC correction factor')
        plt.grid(True)
        plt.show()

def test_gpa_calibration():
    expt = Experiment(init_gpa=True)

    # test calibration on GPA-FHDO capable of driving a full current load
    # if false, just test over a narrow range
    full_current = True

    if full_current:
        expt.gradb.calibrate(channels=[0], max_current=0.7, num_calibration_points=30, averages=5, settle_time=0.005, poly_degree=5)
        expt.gradb.calibrate(channels=[0], max_current=0.7, num_calibration_points=30, averages=1, test_cal=True)
    else:
        expt.gradb.calibrate(channels=[0], max_current=0.05, num_calibration_points=30, averages=5, settle_time=0.005, poly_degree=2)
        expt.gradb.calibrate(channels=[0], max_current=0.05, num_calibration_points=30, averages=1, test_cal=True)

def test_lo_change():
    expt = Experiment(auto_leds=False)
    expt.add_flodict({'tx0': ( np.array([1]), np.array([0.5]) )})
    expt.compile()
    expt.set_lo_freq(2)
    expt.compile()
    expt.run()
    # expt.close_server(only_if_sim=True)

if __name__ == "__main__":
    print("No tests are run.")
    if False:
        test_rx_scaling(lo_freq=0.5,
                        rf_amp=1,
                        rf_steps=False,
                        rx_time=300,
                        rx_padding=10,
                        # rx_periods = np.ones(10, dtype=int)*30,
                        # rx_periods=np.arange(4, 400, 1),
                        # rx_periods=np.arange(4, 40, 1),
                        rx_periods=np.arange(4, 100, 1),
                        # rx_periods=np.arange(10, 400, 1),
                        # rx_periods=np.array([5, 10, 20, 50, 100, 200, 500]),
                        plot_rx=False)

    if False:
        test_gpa_calibration()

    if False:
        test_lo_change()



"""
Main file to run MaRGE
"""
import os
import sys
from PyQt5.QtWidgets import QApplication

print("****************************************************************************************")
print("Graphical User Interface for MaRCoS                                                    *")
print("Dr. J.M. Algarín, mriLab @ i3M, CSIC, Spain                                            *")
print("https://www.i3m-stim.i3m.upv.es/research/magnetic-resonance-imaging-laboratory-mrilab/ *")
print("https://github.com/mriLab-i3M/MaRGE                                                    *")
print("****************************************************************************************")

# Add folders
if not os.path.exists('experiments/parameterization'):
    os.makedirs('experiments/parameterization')
if not os.path.exists('calibration'):
    os.makedirs('calibration')
if not os.path.exists('protocols'):
    os.makedirs('protocols')

from marge.controller.controller_session import SessionController


# Run the gui
app = QApplication(sys.argv)
gui = SessionController()
gui.show()
sys.exit(app.exec_())

# Avant de quitter l'application :
console_controller.close_log()



import threading
import numpy as np
from marge.widgets.widget_preprocessing import PreProcessingTabWidget


class PreProcessingTabController(PreProcessingTabWidget):
    """
    Controller class for the pre-processing tab widget.

    Inherits from PreProcessingTabWidget.

    Attributes:
        partial_reconstruction_button (QPushButton): QPushButton for applying partial reconstruction.
        image_cosbell_button (QPushButton): QPushButton for applying Cosbell filter.
        phase_center_button (QPushButton): QPushButton for getting the phase center.
        image_padding_button (QPushButton): QPushButton for applying zero padding.
        new_fov_button (QPushButton): QPushButton for changing the field of view.
    """

    def __init__(self, *args, **kwargs):
        """
        Initialize the PreProcessingTabController.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        super(PreProcessingTabController, self).__init__(*args, **kwargs)

        # Connect the button click signal to the corresponding methods
        self.scan_button.clicked.connect(self.selectScans)
        self.image_cosbell_button.clicked.connect(self.cosbellFilter)
        self.image_padding_button.clicked.connect(self.zeroPadding)
        self.new_fov_button.clicked.connect(self.fovShifting)

    def selectScans(self):
        # Define method to get the indexes
        def parse_indexes(index_string, max_index):
            if index_string.strip().lower() == "all":
                return list(range(max_index))

            indexes = []

            for part in index_string.split(','):
                part = part.strip()
                if ':' in part:
                    start, end = part.split(':')
                    start = int(start.strip())
                    end = int(end.strip())
                    indexes.extend(range(start, end))
                else:
                    indexes.append(int(part))

            return indexes

        # Get the mat data
        mat_data = self.main.toolbar_image.mat_data

        # Get the desired scans
        n_scans = mat_data['nScans'][0][0]
        scans = self.scan_field.text()
        print(" Selected scans: " + str(scans))
        scans = parse_indexes(scans, n_scans)

        # Generate artifitial data_full for RareDoubleImage
        if mat_data['seqName'] == 'RareDoubleImage':
            data_odd = mat_data['data_full_odd_echoes']
            data_even = mat_data['data_full_even_echoes']
            s_scans, n_sl, n_ph, n_rd = np.shape(data_odd)
            n_points = np.reshape(mat_data['nPoints'], -1)

            # Process odd data
            data_odd_temp = np.zeros((len(scans), n_points[2], n_points[1], n_points[0]), dtype=complex)
            data_odd_temp[:, 0:n_sl, :, :] = data_odd[scans, :, :, :]
            data_odd = np.average(data_odd_temp, axis=0)
            img_odd = np.fft.ifftshift(np.fft.ifftn(np.fft.fftshift(data_odd)))

            # Process even data
            data_even_temp = np.zeros((len(scans), n_points[2], n_points[1], n_points[0]), dtype=complex)
            data_even_temp[:, 0:n_sl, :, :] = data_even[scans, :, :, :]
            data_even = np.average(data_even_temp, axis=0)
            img_even = np.fft.ifftshift(np.fft.ifftn(np.fft.fftshift(data_even)))

            # Get average k-space from odd and even image
            img = (np.abs(img_odd) + np.abs(img_even)) / 2
            data = np.fft.fftshift(np.fft.fftn(np.fft.ifftshift(img)))
        else:
            # Get the selected scans from dataFull and average
            data_full = mat_data['dataFull'][scans, :, :, :]
            s_scans, n_sl, n_ph, n_rd = np.shape(data_full)
            n_points = np.reshape(mat_data['nPoints'], -1)
            data_temp = np.zeros((len(scans), n_points[2], n_points[1], n_points[0]), dtype=complex)
            data_temp[:, 0:n_sl, :, :] = data_full
            data_temp = np.average(data_temp, axis=0)
            data = np.reshape(data_temp, (1, n_points[0] * n_points[1] * n_points[2]))

        # Input the resulting data into the k-space
        self.main.toolbar_image.k_space_raw[:, 3] = np.reshape(data, -1)
        self.main.toolbar_image.k_space = np.reshape(self.main.toolbar_image.k_space_raw[:, 3], n_points[-1::-1])

        # Update the main matrix of the image view widget with the cosbell data
        self.main.image_view_widget.main_matrix = self.main.toolbar_image.k_space.copy()

        # Add new item to the history list
        self.main.history_list.addNewItem(stamp="Selected scans",
                                          image=self.main.image_view_widget.main_matrix,
                                          orientation=self.main.toolbar_image.mat_data['axesOrientation'][0],
                                          operation="Scans: " + str(scans),
                                          space="k",
                                          image_key=self.main.image_view_widget.image_key)


    def cosbellFilter(self):
        """
        Apply the Cosbell filter operation using threading.

        Starts a new thread to execute the runCosbellFilter method.
        """

        thread = threading.Thread(target=self.runCosbellFilter)
        thread.start()

    def runCosbellFilter(self):
        """
        Run the Cosbell filter operation.

        Retrieves the necessary parameters and performs the Cosbell filtering on the loaded image.
        Updates the main matrix of the image view widget with the filtered data, adds the operation to the history widget,
        and updates the operations history.
        """
        # Cosbell filter order from the text field
        cosbell_order = float(self.cosbell_order_field.text())

        # Get the mat data from the loaded .mat file in the main toolbar controller
        mat_data = self.main.toolbar_image.mat_data
        sampled = self.main.toolbar_image.k_space_raw
        data = self.main.image_view_widget.main_matrix.copy()
        nPoints = np.reshape(mat_data['nPoints'], -1)

        # Check which checkboxes are selected
        theta = None
        text = "Cosbell -"
        if self.readout_checkbox.isChecked():
            k = np.reshape(sampled[:, 0], nPoints[-1::-1])
            kmax = np.max(np.abs(k[:]))
            text += ' RD,'
            theta = k / kmax
            data *= (np.cos(theta * (np.pi / 2)) ** cosbell_order)
        if self.phase_checkbox.isChecked():
            k = np.reshape(sampled[:, 1], nPoints[-1::-1])
            kmax = np.max(np.abs(k[:]))
            text += ' PH,'
            theta = k / kmax
            data *= (np.cos(theta * (np.pi / 2)) ** cosbell_order)
        if self.slice_checkbox.isChecked():
            k = np.reshape(sampled[:, 2], nPoints[-1::-1])
            kmax = np.max(np.abs(k[:]))
            text += ' SL,'
            theta = k / kmax
            data *= (np.cos(theta * (np.pi / 2)) ** cosbell_order)

        # Update the main matrix of the image view widget with the cosbell data
        self.main.image_view_widget.main_matrix = data.copy()

        # Add new item to the history list
        self.main.history_list.addNewItem(stamp="Cosbell",
                                          image=data,
                                          orientation=self.main.toolbar_image.mat_data['axesOrientation'][0],
                                          operation=text + " Order: " + str(cosbell_order),
                                          space="k",
                                          image_key=self.main.image_view_widget.image_key)

    def zeroPadding(self):
        """
        Apply the zero-padding operation using threading.

        Starts a new thread to execute the runZeroPadding method.
        """

        thread = threading.Thread(target=self.runZeroPadding)
        thread.start()

    def runZeroPadding(self):
        """
        Run the zero-padding operation.

        Retrieves the necessary parameters and performs the zero-padding on the loaded image.
        Updates the main matrix of the image view widget with the padded image, adds the operation to the history
        widget, and updates the operations history.
        """
        # Get the k_space data and its shape
        k_space = self.main.image_view_widget.main_matrix.copy()
        shape_0 = k_space.shape

        # Determine the new shape after zero-padding
        matrix_size = self.zero_padding_order_field.text().split(',')
        n_rd = int(matrix_size[0])
        n_ph = int(matrix_size[1])
        n_sl = int(matrix_size[2])
        shape_1 = n_sl, n_ph, n_rd

        # Create an image matrix filled with zeros
        image_matrix = np.zeros(shape_1, dtype=complex)

        # Calculate the centering offsets for each dimension
        offset_0 = (shape_1[0] - shape_0[0]) // 2
        offset_1 = (shape_1[1] - shape_0[1]) // 2
        offset_2 = (shape_1[2] - shape_0[2]) // 2

        # Calculate the start and end indices to center the k_space within the new image_matrix
        new_start_0 = offset_0 if offset_0 >= 0 else 0
        new_start_1 = offset_1 if offset_1 >= 0 else 0
        new_start_2 = offset_2 if offset_2 >= 0 else 0
        new_end_0 = new_start_0 + shape_0[0] if offset_0 > 0 else shape_1[0]
        new_end_1 = new_start_1 + shape_0[1] if offset_1 > 0 else shape_1[1]
        new_end_2 = new_start_2 + shape_0[2] if offset_2 > 0 else shape_1[2]

        # Calculate the start and end indices of old matrix
        old_start_0 = 0 if offset_0 >= 0 else -offset_0
        old_start_1 = 0 if offset_1 >= 0 else -offset_1
        old_start_2 = 0 if offset_2 >= 0 else -offset_2
        old_end_0 = shape_0[0] if offset_0 >=0 else old_start_0 + shape_1[0]
        old_end_1 = shape_0[1] if offset_1 >=0 else old_start_1 + shape_1[1]
        old_end_2 = shape_0[2] if offset_2 >=0 else old_start_2 + shape_1[2]

        # Copy the k_space into the image_matrix at the center
        image_matrix[new_start_0:new_end_0, new_start_1:new_end_1, new_start_2:new_end_2] = k_space[
                                                                                            old_start_0:old_end_0,
                                                                                            old_start_1:old_end_1,
                                                                                            old_start_2:old_end_2
                                                                                            ]

        # Update the main matrix of the image view widget with the padded image
        self.main.image_view_widget.main_matrix = image_matrix.copy()

        # Add new item to the history list
        self.main.history_list.addNewItem(stamp="Zero Padding",
                                          image=self.main.image_view_widget.main_matrix,
                                          orientation=self.main.toolbar_image.mat_data['axesOrientation'][0],
                                          operation="Zero Padding - RD: " + str(n_rd) + ", PH: "
                                                    + str(n_ph) + ", SL: " + str(n_sl),
                                          space="k",
                                          image_key=self.main.image_view_widget.image_key)

    def fovShifting(self):
        """
        Perform the FOV change operation using threading.

        Starts a new thread to execute the runFovShifting method.
        """
        thread = threading.Thread(target=self.runFovShifting)
        thread.start()

    def runFovShifting(self):
        """
        Run the FOV change operation.

        Retrieves the necessary parameters and performs the FOV change on the loaded image.
        Updates the main matrix of the image view widget with the new FOV image, adds the operation to the history
        widget, and updates the operations history.
        """
        # Get the k_space data and its shape
        k = self.main.toolbar_image.k_space_raw.copy()
        nPoints = self.main.toolbar_image.nPoints

        # Factors for FOV change from the text field
        factors = self.change_fov_field.text().split(',')

        # Extract the k_space components
        krd = k[:, 0]
        kph = k[:, 1]
        ksl = k[:, 2]
        k = np.column_stack((krd, kph, ksl))

        # Convert factors to spatial shifts
        delta_rd = (float(factors[0])) * 1e-3
        delta_ph = (float(factors[1])) * 1e-3
        delta_sl = (float(factors[2])) * 1e-3
        delta_r = np.array([delta_rd, delta_ph, delta_sl])

        # Calculate the phase shift using the spatial shifts
        phi = np.exp(-1j * 2 * np.pi * k @ np.reshape(delta_r, (3, 1)))
        self.main.image_view_widget.main_matrix *= np.reshape(phi, nPoints[-1::-1])

        # Add the "New FOV" operation to the history widget with a timestamp
        self.main.history_list.addNewItem(stamp="FOV shift",
                                          image=self.main.image_view_widget.main_matrix,
                                          orientation=self.main.toolbar_image.mat_data['axesOrientation'][0],
                                          operation="Fov shift",
                                          space="k",
                                          image_key=self.main.image_view_widget.image_key)



#!/usr/bin/env python3
#
# Base definitions and functions used for marga testing/simulation

import sys, os, subprocess, warnings, socket, unittest, time
import numpy as np
import matplotlib.pyplot as plt

import server_comms as sc

import marcompile as mc
import experiment as exp

import pdb
st = pdb.set_trace
ip_address = "localhost"
port = 11111

# simulation configuration
marga_sim_path = os.path.join("..", "marga")
marga_sim_csv = os.path.join("/tmp", "marga_sim.csv")

# Set to True to debug with GTKWave -- just do one test at a time!
marga_sim_fst_dump = False
marga_sim_fst = os.path.join("/tmp", "marga_sim.fst")

# Arguments for compare_csv when running gradient tests
fhd_config = {
    'initial_bufs': np.array([
        # see marga.sv, gradient control lines (lines 186-190, 05.02.2021)
        # strobe for both LSB and LSB, reset_n = 1, spi div = 10, grad board select (1 = ocra1, 2 = gpa-fhdo)
        (1 << 9) | (1 << 8) | (10 << 2) | 2,
        0, 0,
        0, 0,
        0, 0, 0, 0,
        0, 0, 0, 0, 0, 0,
        0, 0], dtype=np.uint16),
    'latencies': np.array([
        0, 276, 276, # grad latencies match SPI div
        0, 0, # rx
        0, 0, 0, 0, # tx
        0, 0, 0, 0, 0, 0, # lo phase
        0, 0 # gates and LEDs, RX config
    ], dtype=np.uint16)}

oc1_config = {
    'initial_bufs': np.array([
        # see marga.sv, gradient control lines (lines 186-190, 05.02.2021)
        # strobe for both LSB and LSB, reset_n = 1, spi div = 10, grad board select (1 = ocra1, 2 = gpa-fhdo)
        (1 << 9) | (1 << 8) | (10 << 2) | 1,
        0, 0,
        0, 0,
        0, 0, 0, 0,
        0, 0, 0, 0, 0, 0,
        0, 0], dtype=np.uint16),
    'latencies': np.array([
        0, 268, 268, # grad latencies match SPI div
        0, 0, # rx
        0, 0, 0, 0, # tx
        0, 0, 0, 0, 0, 0, # lo phase
        0, 0 # gates and LEDs, RX config
    ], dtype=np.uint16)}

gb_orig = None
gb_changed = False

def set_grad_board(gb):
    global gb_orig, gb_changed
    if not gb_changed:
        gb_orig = mc.grad_board
    mc.grad_board = gb
    exp.grad_board = gb
    gb_changed = True

def restore_grad_board():
    global gb_orig, gb_changed
    mc.grad_board = gb_orig
    exp.grad_board = gb_orig
    gb_changed = False

def compare_csv(fname, sock, proc,
                 initial_bufs=np.zeros(mc.MARGA_BUFS, dtype=np.uint16),
                 latencies=np.zeros(mc.MARGA_BUFS, dtype=np.uint32),
                 self_ref=True # use the CSV source file as the reference file to compare the output with
                 ):

    source_csv = os.path.join("csvs", fname + ".csv")
    lc = mc.csv2bin(source_csv,
                    quick_start=False, initial_bufs=initial_bufs, latencies=latencies)
    data = np.array(lc, dtype=np.uint32)

    # run simulation
    rx_data, msgs = sc.command({'run_seq': data.tobytes()} , sock)

    # halt simulation
    sc.send_packet(sc.construct_packet({}, 0, command=sc.close_server_pkt), sock)
    sock.close()
    proc.wait(1) # wait a short time for simulator to close

    # compare resultant CSV with the reference
    if self_ref:
        rdata = np.loadtxt(source_csv, skiprows=1, delimiter=',', comments='#').astype(np.uint32)
        sdata = np.loadtxt(marga_sim_csv, skiprows=1, delimiter=',', comments='#').astype(np.uint32)

        rdata[1:,0] -= rdata[1,0] # subtract off initial offset time
        sdata[1:,0] -= sdata[1,0] # subtract off initial offset time

        return rdata.tolist(), sdata.tolist()
    else:
        ref_csv = os.path.join("csvs", "ref_" + fname + ".csv")
        with open(ref_csv, "r") as ref:
            refl = ref.read().splitlines()
        with open(marga_sim_csv, "r") as sim:
            siml = sim.read().splitlines()
        return refl, siml

def compare_dict(source_dict, ref_fname, sock, proc,
                 initial_bufs=np.zeros(mc.MARGA_BUFS, dtype=np.uint16),
                 latencies=np.zeros(mc.MARGA_BUFS, dtype=np.uint32),
                 ignore_start_delay=True
                 ):

    lc = mc.dict2bin(source_dict, initial_bufs=initial_bufs, latencies=latencies)
    data = np.array(lc, dtype=np.uint32)

    # run simulation
    rx_data, msgs = sc.command({'run_seq': data.tobytes()} , sock)

    # halt simulation
    sc.send_packet(sc.construct_packet({}, 0, command=sc.close_server_pkt), sock)
    sock.close()
    proc.wait(1) # wait a short time for simulator to close

    ref_csv = os.path.join("csvs", ref_fname + ".csv")
    with open(ref_csv, "r") as ref:
        refl = ref.read().splitlines()
    with open(marga_sim_csv, "r") as sim:
        siml = sim.read().splitlines()
    # return refl, siml

    ref_csv = os.path.join("csvs", ref_fname + ".csv")
    if ignore_start_delay:
        rdata = np.loadtxt(ref_csv, skiprows=1, delimiter=',', comments='#').astype(np.uint32)
        sdata = np.loadtxt(marga_sim_csv, skiprows=1, delimiter=',', comments='#').astype(np.uint32)

        rdata[1:,0] -= rdata[1,0] # subtract off initial offset time
        sdata[1:,0] -= sdata[1,0] # subtract off initial offset time

        return rdata.tolist(), sdata.tolist()
    else:
        with open(ref_csv, "r") as ref:
            refl = ref.read().splitlines()
        with open(marga_sim_csv, "r") as sim:
            siml = sim.read().splitlines()
        return refl, siml

def expt_run(e):
    """ Function for customising how compare_expt_dict() runs Experiments; e.g. for testing different Experiment methods etc
    (see test_lo_change_expt in test_marga_model.py for an example)"""
    rx_data, msgs = e.run()
    return rx_data, msgs

def compare_expt_dict(source_dict, ref_fname, sock, proc,
                      # initial_bufs=np.zeros(mc.MARGA_BUFS, dtype=np.uint16),
                      # latencies=np.zeros(mc.MARGA_BUFS, dtype=np.uint32),
                      ignore_start_delay=True,
                      run_fn=expt_run,
                      **kwargs):
    """Arguments the same as for compare_dict(), except that the source
    dictionary is in floating-point units, and the kwargs are passed
    to the Experiment class constructor. Note that the initial_bufs
    and latencies are supplied to the Experiment class from the
    classes in grad_board.py.
    """

    e = exp.Experiment(prev_socket=sock, seq_dict=source_dict, **kwargs)

    # run simulation
    rx_data, msgs = run_fn(e)

    # halt simulation
    sc.send_packet(sc.construct_packet({}, 0, command=sc.close_server_pkt), sock)
    sock.close()
    proc.wait(1) # wait a short time for simulator to close

    ref_csv = os.path.join("csvs", ref_fname + ".csv")
    with open(ref_csv, "r") as ref:
        refl = ref.read().splitlines()
    with open(marga_sim_csv, "r") as sim:
        siml = sim.read().splitlines()
    # return refl, siml

    ref_csv = os.path.join("csvs", ref_fname + ".csv")
    if ignore_start_delay:
        rdata = np.loadtxt(ref_csv, skiprows=1, delimiter=',', comments='#').astype(np.uint32)
        sdata = np.loadtxt(marga_sim_csv, skiprows=1, delimiter=',', comments='#').astype(np.uint32)

        rdata[1:,0] -= rdata[1,0] # subtract off initial offset time
        sdata[1:,0] -= sdata[1,0] # subtract off initial offset time

        return rdata.tolist(), sdata.tolist()
    else:
        with open(ref_csv, "r") as ref:
            refl = ref.read().splitlines()
        with open(marga_sim_csv, "r") as sim:
            siml = sim.read().splitlines()
        return refl, siml



"""
@author:    José Miguel Algarín
@email:     josalggui@i3m.upv.es
@affiliation:MRILab, i3M, CSIC, Valencia, Spain
"""
from PyQt5.QtWidgets import QListWidget, QSizePolicy


class OutputListWidget(QListWidget):
    def __init__(self, parent, *args, **kwargs):
        super(OutputListWidget, self).__init__(*args, **kwargs)
        self.main = parent
        self.setMaximumHeight(200)



#/usr/bin/env python3
# Machine code defines and functions for marga.
#
# Functions should be fast, without any floating-point arithmetic -
# that should be handled at a higher level.

import numpy as np

class MarUserWarning(UserWarning):
    pass

class MarCompileWarning(MarUserWarning):
    pass

class MarRemovedInstructionWarning(MarCompileWarning):
    pass

class MarGradWarning(MarUserWarning):
    pass

class MarServerWarning(MarUserWarning):
    pass

INOP = 0x0
IFINISH = 0x1
IWAIT = 0x2
ITRIG = 0x3
ITRIGFOREVER=0x4
IDATA = 0x80

GRAD_CTRL = 0
GRAD_LSB = 1
GRAD_MSB = 2
RX0_RATE = 3
RX1_RATE = 4
TX0_I = 5
TX0_Q = 6
TX1_I = 7
TX1_Q = 8
DDS0_PHASE_LSB = 9
DDS0_PHASE_MSB = 10
DDS1_PHASE_LSB = 11
DDS1_PHASE_MSB = 12
DDS2_PHASE_LSB = 13
DDS2_PHASE_MSB = 14
GATES_LEDS = 15
RX_CTRL = 16
MARGA_BUFS = RX_CTRL + 1

STATE_IDLE = 0
STATE_PREPARE = 1
STATE_RUN = 2
STATE_COUNTDOWN = 3
STATE_TRIG = 4
STATE_TRIG_FOREVER = 5
STATE_HALT = 8

COUNTER_MAX = 0xffffff

CIC_STAGES = 6 # N: number of CIC stages in the RX CICs
# diff_delay = 1 # M: differential delay in comb section of CICs
CIC_RATE_DATAWIDTH = 12 # 12-bit rate/data bus, 2-bit address
CIC_FASTEST_RATE, CIC_SLOWEST_RATE = 4, 4095 # CIC core settings

def insta(instr: np.uint8, data: np.uint32):
    """ Instruction A: FSM control """
    assert instr in [INOP, IFINISH, IWAIT, ITRIG, ITRIGFOREVER], "Unknown instruction"
    assert (data & COUNTER_MAX) == (data & 0xffffffff), "Data out of range"
    return (instr << 24) | (data & 0xffffff)

def instb(tgt: np.uint8, delay: np.uint8, data: np.uint16):
    """ Instruction B: timed buffered data """
    assert tgt <= 24, "Unknown target buffer"
    assert 0 <= delay <= 255, "Delay out of range"
    assert (np.uint32(data) & 0xffff) == (np.uint32(data) & 0xffffffff), "Data out of range"
    return (IDATA << 24) | ( (tgt & 0x7f) << 24 ) | (delay << 16) | np.uint32(data)



import numpy as np

# Config file for Physio MRI scanner at MRILab, i3M, CSIC, Spain.

# Note: I write Ocra1 Units as o.u.
# Ocra1 gain = 10 V/o.u.
# AE Techrom transductance. Unbalanced input: 5 A/V. Balanced input: 10 A/V.
# From Ocra1 to current: 50 A/o.u.
# X axis: 25 mT/m/o.u., 0.5 mT/m/A, 2.5 mT/m/V
# Y axis: 40 mT/m/o.u., 0.8 mT/m/A, 4.0 mT/m/V
# Z axis: 35 mT/m/o.u., 0.7 mT/m/A, 3.5 mT/m/V

#gFactor = np.array([0.025, 0.040, 0.035])*2 # (X, Y, Z) in T/m/o.u. GX VIEJO
gFactor = np.array([0.035, 0.040, 0.035])*2 # (X, Y, Z) in T/m/o.u. GX NUEVO
max_grad = 80  # mT/m
max_slew_rate = 50e-3  # mT/m/ms
grad_raster_time = 25e-6  # s
grad_rise_time = 400e-6 # s, time for gradient ramps
grad_steps = 16 # steps to gradient ramps
gammaB = 42.57747892e6 # Hz/T, Gyromagnetic ratio
blkTime = 15 # us, blanking time of Barthel's RFPA
gradDelay = 9 # Gradient amplifier delay (us)
oversamplingFactor = 5 # Rx oversampling
maxRdPoints = 2**17 # Maximum number of points to be acquired by the red pitaya
maxOrders = 2**14*6 # Maximun number of orders to be processed by the red pitaya
deadTime = 400 # us, RF coil dead time
b1Efficiency = np.pi/(1.0*15) # rads / (a.u. * us)
larmorFreq = 3.053482675254997 # MHz
cic_delay_points = 3 # to account for signal delay from red pitaya due to cic filter
addRdPoints = 10 # to account for wrong first points after decimation (50)
scanner_name = "Demo"
antenna_dict = {}
reference_time = 100  # us
fov = [18.0, 16.0, 16.0] # cm
dfov = [0.0, 0.0, 0.0] # mm

# To change the original reference system. Check in MaRGE wiki for more info.
rotations = []
dfovs = []
fovs = []

bash_path = "gnome-terminal"
rp_ip_list = []
rp_version = "rp-122"
adcFactor = 13.788
lnaGain = 45 # dB
temperature = 293 # k
shimming_factor = 1e-4
rf_min_gain = 35
rf_max_gain = 50


# Arduinos
ard_sn_autotuning = '4423431343435131E180'
ard_sn_interlock = '242353133363518050E1'
ard_sn_attenuator = '242353133363518050E2'


#!/usr/bin/env python3
#
# loopback test using ocra-pulseq
#

import numpy as np
import matplotlib.pyplot as plt
import pdb
import experiment as ex
from pulseq_assembler import PSAssembler
st = pdb.set_trace

if __name__ == "__main__":
    lo_freq = 2.1 # MHz
    tx_t = 1.001 # us
    rx_t = 0.497
    clk_t = 0.007
    num_grad_channels = 3
    grad_interval = 10.003 # us between [num_grad_channels] channel updates

    gamma = 42570000 # Hz/T

    # value for tabletopMRI  gradient coil
    B_per_m_per_current = 0.02 # T/m/A, approximate value for tabletop gradient coil

    # values for gpa fhdo
    gpa_current_per_volt = 3.75 # A/V, theoretical value for gpa fhdo without 2.5 V offset!
    max_dac_voltage = 2.5

    # values for ocra1
    # max_dac_voltage = 5
    # gpa_current_per_volt = 3.75 # A/V, fill in value of gradient power amplifier here!

    max_Hz_per_m = max_dac_voltage * gpa_current_per_volt * B_per_m_per_current * gamma	
    # grad_max = max_Hz_per_m # factor used to normalize gradient amplitude, should be max value of the gpa used!	
    grad_max = 16 # unrealistic value used only for loopback test

    rf_amp_max = 10 # factor used to normalize RF amplitude, should be max value of system used!
    ps = PSAssembler(rf_center=lo_freq*1e6,
        # how many Hz the max amplitude of the RF will produce; i.e. smaller causes bigger RF V to compensate
        rf_amp_max=rf_amp_max,
        grad_max=grad_max,
        clk_t=clk_t,
        tx_t=tx_t,
        grad_t=grad_interval)
    tx_arr, grad_arr, cb, params = ps.assemble('test_gpa_fhdo.seq', byte_format=False)

    # Temporary hack, until next ocra-pulseq update
    if 'rx_t' not in params:
        params['rx_t'] = rx_t
    
    exp = ex.Experiment(samples=params['readout_number'], 
        lo_freq=lo_freq,
        tx_t=tx_t,
		rx_t=params['rx_t'],
        grad_channels=num_grad_channels,
        grad_t=grad_interval/num_grad_channels,
        assert_errors=False)

    exp.define_instructions(cb)
    x = np.linspace(0,2*np.pi, 100)
    ramp_sine = np.sin(2*x)
    exp.add_tx(tx_arr)
    exp.add_grad(grad_arr)

    exp.gradb.calibrate(max_current = 2,
        num_calibration_points=10,
        gpa_current_per_volt=gpa_current_per_volt,
        averages=4,
        plot=False)

    current = 0.1 # ampere
    # set all channels back to 0.1 A
    for ch in range(num_grad_channels):
        dac_code = exp.gradb.ampere_to_dac_code(current)
        dac_code = exp.gradb.calculate_corrected_dac_code(ch,dac_code)
        exp.gradb.write_dac(ch, dac_code)

    wait = input('All gradient channels are set to {:f} A. Press Enter to continue.'.format(current))

    data = exp.run() # Comment out this line to avoid running on the hardware

    # set all channels back to 0 A
    for ch in range(num_grad_channels):
        dac_code = exp.gradb.ampere_to_dac_code(0)
        dac_code = exp.gradb.calculate_corrected_dac_code(ch,dac_code)
        exp.gradb.write_dac(ch, dac_code)        

    # st()



#!/usr/bin/env python3

import numpy as np
import matplotlib.pyplot as plt
import experiment as ex
from local_config import grad_board

import pdb
st = pdb.set_trace

def trapezoid(plateau_a, total_t, ramp_t, ramp_pts, total_t_end_to_end=True, base_a=0):
    """Helper function that just generates a Numpy array starting at time
    0 and ramping down at time total_t, containing a trapezoid going from a
    level base_a to plateau_a, with a rising ramp of duration ramp_t and
    sampling period ramp_ts."""

    # ramp_pts = int( np.ceil(ramp_t/ramp_ts) ) + 1
    rise_ramp_times = np.linspace(0, ramp_t, ramp_pts)
    rise_ramp = np.linspace(base_a, plateau_a, ramp_pts)

    # [1: ] because the first element of descent will be repeated
    descent_t = total_t - ramp_t if total_t_end_to_end else total_t
    t = np.hstack([rise_ramp_times, rise_ramp_times[:-1] + descent_t])
    a = np.hstack([rise_ramp, np.flip(rise_ramp)[1:]])
    return t, a

def trap_cent(centre_t, plateau_a, trap_t, ramp_t, ramp_pts, base_a=0):
    """Like trapezoid, except it generates a trapezoid shape around a centre
    time, with a well-defined area given by its amplitude (plateau_a)
    times its time (trap_t), which is defined from the start of the
    ramp-up to the start of the ramp-down, or (equivalently) from the
    centre of the ramp-up to the centre of the ramp-down. All other
    parameters are as for trapezoid()."""
    t, a = trapezoid(plateau_a, trap_t, ramp_t, ramp_pts, False, base_a)
    return t + centre_t - (trap_t + ramp_t)/2, a

def grad_echo(trs=21, plot_rx=False, init_gpa=False, plot_sequence=False,
              dbg_sc=0.5, # set to 0 to avoid 2nd RF debugging pulse, otherwise amp between 0 or 1
              lo_freq=0.1, # MHz
              rf_amp=1, # 1 = full-scale

              slice_amp=0.4, # 1 = gradient full-scale
              phase_amp=0.3, # 1 = gradient full-scale
              readout_amp=0.8, # 1 = gradient full-scale
              rf_duration=50,
              trap_ramp_duration=50, # us, ramp-up/down time
              trap_ramp_pts=5, # how many points to subdivide ramp into
              phase_delay=100, # how long after RF end before starting phase ramp-up
              phase_duration=200, # length of phase plateau
              tr_wait=100, # delay after end of RX before start of next TR

              rx_period=10/3 # us, 3.333us, 300 kHz rate
              ):
    ## All times are in the context of a single TR, starting at time 0

    phase_amps = np.linspace(phase_amp, -phase_amp, trs)

    rf_tstart = 100 # us
    rf_tend = rf_tstart + rf_duration # us

    slice_tstart = rf_tstart - trap_ramp_duration
    slice_duration = (rf_tend - rf_tstart) + 2*trap_ramp_duration # includes rise, plateau and fall
    phase_tstart = rf_tend + phase_delay
    readout_tstart = phase_tstart
    readout_duration = phase_duration*2

    rx_tstart = readout_tstart + trap_ramp_duration # us
    rx_tend = readout_tstart + readout_duration - trap_ramp_duration # us

    tx_gate_pre = 2 # us, time to start the TX gate before the RF pulse begins
    tx_gate_post = 1 # us, time to keep the TX gate on after the RF pulse ends

    tr_total_time = readout_tstart + readout_duration + tr_wait + 7000 # start-finish TR time

    def grad_echo_tr(tstart, pamp):
        gvxt, gvxa = trapezoid(slice_amp, slice_duration, trap_ramp_duration, trap_ramp_pts)
        gvyt, gvya = trapezoid(pamp, phase_duration, trap_ramp_duration, trap_ramp_pts)

        gvzt1 = trapezoid(readout_amp, readout_duration/2, trap_ramp_duration, trap_ramp_pts)
        gvzt2 = trapezoid(-readout_amp, readout_duration/2, trap_ramp_duration, trap_ramp_pts)
        gvzt = np.hstack([gvzt1[0], gvzt2[0] + readout_duration/2])
        gvza = np.hstack([gvzt1[1], gvzt2[1]])

        rx_tcentre = (rx_tstart + rx_tend) / 2

        value_dict = {
            # second tx0 pulse purely for loopback debugging
            'tx0': ( np.array([rf_tstart, rf_tend,   rx_tcentre - 10, rx_tcentre + 10]) + tstart,
                     np.array([rf_amp,0,  dbg_sc*(1 + 0.5j),0]) ),

            'tx1': ( np.array([rx_tstart + 15, rx_tend - 15]) + tstart, np.array([dbg_sc * pamp * (1 + 0.5j), 0]) ),
            'grad_vx': ( gvxt + tstart + slice_tstart, gvxa ),
            'grad_vy': ( gvyt + tstart + phase_tstart, gvya),
            'grad_vz': ( gvzt + tstart + readout_tstart, gvza),
            'rx0_en': ( np.array([rx_tstart, rx_tend]) + tstart, np.array([1, 0]) ),
            'rx1_en': ( np.array([rx_tstart, rx_tend]) + tstart, np.array([1, 0]) ), # acquire on RX1 for example too
            'rx_gate': ( np.array([rx_tstart, rx_tend]) + tstart, np.array([1, 0]) ),
            'tx_gate': ( np.array([rf_tstart - tx_gate_pre, rf_tend + tx_gate_post]) + tstart, np.array([1, 0]) )
        }

        return value_dict

    expt = ex.Experiment(lo_freq=lo_freq, rx_t=rx_period, init_gpa=init_gpa, gpa_fhdo_offset_time=(1 / 0.2 / 3.1))
    # gpa_fhdo_offset_time in microseconds; offset between channels to
    # avoid parallel updates (default update rate is 0.2 Msps, so
    # 1/0.2 = 5us, 5 / 3.1 gives the offset between channels; extra
    # 0.1 for a safety margin)

    tr_t = 0 # start the first TR at 20us
    for pamp in phase_amps:
        expt.add_flodict( grad_echo_tr( tr_t, pamp) )
        tr_t += tr_total_time

    if plot_sequence:
        expt.plot_sequence()
        plt.show()

    rxd, msgs = expt.run()
    expt.close_server(True)
    expt._s.close() # close socket

    if plot_rx:
        plt.plot( rxd['rx0'].real )
        plt.plot( rxd['rx0'].imag )
        plt.plot( rxd['rx1'].real )
        plt.plot( rxd['rx1'].imag )
        plt.show()

def turbo_spin_echo(plot_rx=False, init_gpa=False, plot_sequence=False,
                    dbg_sc=0.5, # set to 0 to avoid RF debugging pulses in each RX window, otherwise amp between 0 or 1
                    lo_freq=0.2, # MHz
                    rf_amp=1, # 1 = full-scale

                    rf_pi2_duration=50, # us, rf pi/2 pulse length
                    rf_pi_duration=None, # us, rf pi pulse length  - if None then automatically gets set to 2 * rf_pi2_duration

                    # trapezoid properties - shared between all gradients for now
                    trap_ramp_duration=50, # us, ramp-up/down time
                    trap_ramp_pts=5, # how many points to subdivide ramp into

                    # spin-echo properties
                    echos_per_tr=5, # number of spin echoes (180 pulses followed by readouts) to do
                    echo_duration=2000, # us, time from the centre of one echo to centre of the next

                    readout_amp=0.8, # 1 = gradient full-scale
                    readout_duration=500, # us, time in the centre of an echo when the readout occurs
                    rx_period=10/3, # us, 3.333us, 300 kHz rate
                    readout_grad_duration=700, # us, readout trapezoid lengths (mid-ramp-up to mid-ramp-down)
                    # (must at least be longer than readout_duration + trap_ramp_duration)

                    phase_start_amp=0.6, # 1 = gradient full-scale, starting amplitude (by default ramps from +ve to -ve in each echo)
                    phase_grad_duration=150, # us, phase trapezoid lengths (mid-ramp-up to mid-ramp-down)
                    phase_grad_interval=1200, # us, interval between first phase trapezoid and its negative-sign counterpart within a single echo

                    # slice trapezoid timing is the same as phase timing
                    slice_start_amp=0.3, # 1 = gradient full-scale, starting amplitude (by default ramps from +ve to -ve in each TR)

                    tr_pause_duration=3000, # us, length of time to pause from the end of final echo's RX pulse to start of next TR
                    trs=5 # number of TRs
                    ):
    """
    readout gradient: x
    phase gradient: y
    slice/partition gradient: z
    """

    if rf_pi_duration is None:
        rf_pi_duration = 2 * rf_pi2_duration

    phase_amps = np.linspace(phase_start_amp, -phase_start_amp, echos_per_tr)
    slice_amps = np.linspace(slice_start_amp, -slice_start_amp, trs)

    # create appropriate waveforms for each echo, based on start time, echo index and TR index
    # note: echo index is 0 for the first interval (90 pulse until first 180 pulse) thereafter 1, 2 etc between each 180 pulse
    def rf_wf(tstart, echo_idx):
        pi2_phase = 1 # x
        pi_phase = 1j # y
        if echo_idx == 0:
            # do pi/2 pulse, then start first pi pulse
            return np.array([tstart + (echo_duration - rf_pi2_duration)/2, tstart + (echo_duration + rf_pi2_duration)/2,
                             tstart + echo_duration - rf_pi_duration/2]), np.array([pi2_phase, 0, pi_phase]) * rf_amp
        elif echo_idx == echos_per_tr:
            # finish final RF pulse
            return np.array([tstart + rf_pi_duration/2]), np.array([0])
        else:
            # finish last pi pulse, start next pi pulse
            return np.array([tstart + rf_pi_duration/2, tstart + echo_duration - rf_pi_duration/2]), np.array([0, pi_phase]) * rf_amp

    def tx_gate_wf(tstart, echo_idx):
        tx_gate_pre = 2 # us, time to start the TX gate before each RF pulse begins
        tx_gate_post = 1 # us, time to keep the TX gate on after an RF pulse ends

        if echo_idx == 0:
            # do pi/2 pulse, then start first pi pulse
            return np.array([tstart + (echo_duration - rf_pi2_duration)/2 - tx_gate_pre,
                             tstart + (echo_duration + rf_pi2_duration)/2 + tx_gate_post,
                             tstart + echo_duration - rf_pi_duration/2 - tx_gate_pre]), \
                             np.array([1, 0, 1])
        elif echo_idx == echos_per_tr:
            # finish final RF pulse
            return np.array([tstart + rf_pi_duration/2 + tx_gate_post]), np.array([0])
        else:
            # finish last pi pulse, start next pi pulse
            return np.array([tstart + rf_pi_duration/2 + tx_gate_post, tstart + echo_duration - rf_pi_duration/2 - tx_gate_pre]), \
                np.array([0, 1])

    def readout_grad_wf(tstart, echo_idx):
        if echo_idx == 0:
            return trap_cent(tstart + echo_duration*3/4, readout_amp, readout_grad_duration/2,
                             trap_ramp_duration, trap_ramp_pts)
        else:
            return trap_cent(tstart + echo_duration/2, readout_amp, readout_grad_duration,
                             trap_ramp_duration, trap_ramp_pts)

    def readout_wf(tstart, echo_idx):
        if echo_idx != 0:
            return np.array([tstart + (echo_duration - readout_duration)/2, tstart + (echo_duration + readout_duration)/2 ]), np.array([1, 0])
        else:
            return np.array([tstart]), np.array([0]) # keep on zero otherwise

    def phase_grad_wf(tstart, echo_idx):
        t1, a1 = trap_cent(tstart + (echo_duration - phase_grad_interval)/2, phase_amps[echo_idx-1], phase_grad_duration,
                           trap_ramp_duration, trap_ramp_pts)
        t2, a2 = trap_cent(tstart + (echo_duration + phase_grad_interval)/2, -phase_amps[echo_idx-1], phase_grad_duration,
                           trap_ramp_duration, trap_ramp_pts)
        if echo_idx == 0:
            return np.array([tstart]), np.array([0]) # keep on zero otherwise
        elif echo_idx == echos_per_tr: # last echo, don't need 2nd trapezoids
            return t1, a1
        else: # otherwise do both trapezoids
            return np.hstack([t1, t2]), np.hstack([a1, a2])

    def slice_grad_wf(tstart, echo_idx, tr_idx):
        t1, a1 = trap_cent(tstart + (echo_duration - phase_grad_interval)/2, slice_amps[tr_idx], phase_grad_duration,
                           trap_ramp_duration, trap_ramp_pts)
        t2, a2 = trap_cent(tstart + (echo_duration + phase_grad_interval)/2, -slice_amps[tr_idx], phase_grad_duration,
                           trap_ramp_duration, trap_ramp_pts)
        if echo_idx == 0:
            return np.array([tstart]), np.array([0]) # keep on zero otherwise
        elif echo_idx == echos_per_tr: # last echo, don't need 2nd trapezoids
            return t1, a1
        else: # otherwise do both trapezoids
            return np.hstack([t1, t2]), np.hstack([a1, a2])

    tr_total_time = echo_duration * (echos_per_tr + 1) + tr_pause_duration

    expt = ex.Experiment(lo_freq=lo_freq, rx_t=rx_period, init_gpa=init_gpa, gpa_fhdo_offset_time=(1 / 0.2 / 3.1))
    # gpa_fhdo_offset_time in microseconds; offset between channels to
    # avoid parallel updates (default update rate is 0.2 Msps, so
    # 1/0.2 = 5us, 5 / 3.1 gives the offset between channels; extra
    # 0.1 for a safety margin))

    global_t = 0 # start the first TR at this time

    for tr in range(trs):
        for echo in range(echos_per_tr + 1):
            tx_t, tx_a = rf_wf(global_t, echo)
            tx_gate_t, tx_gate_a = tx_gate_wf(global_t, echo)
            readout_t, readout_a = readout_wf(global_t, echo)
            rx_gate_t, rx_gate_a = readout_wf(global_t, echo)

            readout_grad_t, readout_grad_a = readout_grad_wf(global_t, echo)
            phase_grad_t, phase_grad_a = phase_grad_wf(global_t, echo)
            slice_grad_t, slice_grad_a = slice_grad_wf(global_t, echo, tr)

            global_t += echo_duration

            expt.add_flodict({
                'tx0': (tx_t, tx_a),
                'tx1': (tx_t, tx_a),
                'grad_vx': (readout_grad_t, readout_grad_a),
                'grad_vy': (phase_grad_t, phase_grad_a),
                'grad_vz': (slice_grad_t, slice_grad_a),
                'rx0_en': (readout_t, readout_a),
                'rx1_en': (readout_t, readout_a),
                'tx_gate': (tx_gate_t, tx_gate_a),
                'rx_gate': (rx_gate_t, rx_gate_a),
            })

        global_t += tr_pause_duration

    if plot_sequence:
        expt.plot_sequence()
        plt.show()

    rxd, msgs = expt.run()
    expt.close_server(True)
    expt._s.close() # close socket

    if plot_rx:
        plt.plot( rxd['rx0'].real )
        plt.plot( rxd['rx0'].imag )
        plt.plot( rxd['rx1'].real )
        plt.plot( rxd['rx1'].imag )
        plt.show()

def radial(trs=36, plot_rx=False, init_gpa=False, plot_sequence=False):
    ## All times are relative to a single TR, starting at time 0
    lo_freq = 0.2 # MHz
    rf_amp = 0.5 # 1 = full-scale

    G = 0.5 # Gx = G cos (t), Gy = G sin (t)
    angles = np.linspace(0, 2*np.pi, trs) # angle

    gradz_tstart = 0 # us
    grady_tstart = gradz_tstart # us
    rf_tstart = 5 # us
    rf_tend = 50 # us
    rx_tstart = 70 # us
    rx_tend = 180 # us
    rx_period = 3 # us
    tr_total_time = 220 # start-finish TR time
    # tr_total_time = 5000 # start-finish TR time

    tx_gate_pre = 2 # us, time to start the TX gate before the RF pulse begins
    tx_gate_post = 1 # us, time to keep the TX gate on after the RF pulse ends

    def radial_tr(tstart, th):
        dbg_sc=0.01
        gx = G * np.cos(th)
        gy = G * np.sin(th)

        value_dict = {
            # second tx0 pulse and tx1 pulse purely for loopback debugging
            'tx0': ( np.array([rf_tstart, rf_tend,    rx_tstart + 15, rx_tend - 15]),
                     np.array([rf_amp, 0,    dbg_sc * (gx+gy*1j), 0]) ),
            'tx1': ( np.array([rx_tstart + 15, rx_tend - 15]), np.array([dbg_sc * (gx+gy*1j), 0]) ),
            'grad_vz': ( np.array([gradz_tstart]),
                         np.array([gx]) ),
            'grad_vy': ( np.array([grady_tstart]),
                         np.array([gy]) ),
            'rx0_en' : ( np.array([rx_tstart, rx_tend]),
                            np.array([1, 0]) ),
            'rx_gate' : ( np.array([rx_tstart, rx_tend]),
                            np.array([1, 0]) ),
            'tx_gate' : ( np.array([rf_tstart - tx_gate_pre, rf_tend + tx_gate_post]),
                          np.array([1, 0]) )
            }

        for k, v in value_dict.items():
            # v for read, value_dict[k] for write
            value_dict[k] = (v[0] + tstart, v[1])

        return value_dict

    expt = ex.Experiment(lo_freq=lo_freq, rx_t=rx_period, init_gpa=init_gpa, gpa_fhdo_offset_time=(1 / 0.2 / 3.1))
    # gpa_fhdo_offset_time in microseconds; offset between channels to
    # avoid parallel updates (default update rate is 0.2 Msps, so
    # 1/0.2 = 5us, 5 / 3.1 gives the offset between channels; extra
    # 0.1 for a safety margin))

    tr_t = 0 # start the first TR at 0us
    for th in angles:
        expt.add_flodict( radial_tr( tr_t, th ) )
        tr_t += tr_total_time

    if plot_sequence:
        expt.plot_sequence()
        plt.show()

    rxd, msgs = expt.run()
    expt.close_server(True)
    expt._s.close() # close socket

    if plot_rx:
        plt.plot( rxd['rx0'].real )
        plt.plot( rxd['rx0'].imag )
        # plt.plot( rxd['rx1'].real )
        # plt.plot( rxd['rx1'].imag )
        plt.show()

if __name__ == "__main__":
    # import cProfile
    # cProfile.run('test_grad_echo_loop()')
    # for k in range(100):
    # grad_echo(lo_freq=1, trs=1, plot_rx=True, init_gpa=True, dbg_sc=1)
    radial(trs=100, init_gpa=True, plot_rx=True)
    # turbo_spin_echo(trs=2, echos_per_tr=4, plot_sequence=False, tr_pause_duration=100000)

    if False:
        # Stress test: run lots of sequences on the server - should
        # take around a day
        k = 0
        while k < 100000:
            k += 1
            print("TSE, trial {:d}".format(k))
            turbo_spin_echo(lo_freq=0.2,
                            trs=2, echos_per_tr=6,
                            rf_amp=0.7,
                            echo_duration=1000,
                            readout_amp=0.2,
                            readout_duration=300,
                            readout_grad_duration=450,
                            phase_grad_duration=150,
                            phase_grad_interval=600,
                            tr_pause_duration=2000,
                            init_gpa=True, plot_rx=False)



import bm4d
import numpy as np
import nibabel as nib
from skimage.util import view_as_blocks

import scipy as sp
from marge.manager.dicommanager import DICOMImage
import pydicom
from pydicom.dataset import Dataset, FileDataset
from datetime import datetime


def fix_image_orientation(image, axes, orientation='FFS'):
    """
    Adjusts the orientation of a 3D image array to match standard anatomical planes
    (sagittal, coronal, or transversal) and returns the oriented image along with labeling
    and metadata for visualization.

    Args:
        image (np.ndarray): A 3D numpy array representing the image data to be reoriented.
        axes (list[int]): A list of three integers representing the current order of the
                          axes in the image (e.g., [0, 1, 2] for x, y, z).

    Returns:
        output (dict): A dictionary containing the following keys:
            - 'widget': A fixed string "image" or "curve" indicating the type of data for visualization.
            - 'data': The reoriented 3D image array (np.ndarray).
            - 'xLabel': A string representing the label for the x-axis in the visualization.
            - 'yLabel': A string representing the label for the y-axis in the visualization.
            - 'title': A string representing the title of the visualization (e.g., "Sagittal").
        image (np.ndarray): Reoriented 3D image array
        dicom_orientation (list): orientation for dicom file
    """

    # Get axes in strings
    axes_dict = {'x': 0, 'y': 1, 'z': 2}
    axes_keys = list(axes_dict.keys())
    axes_vals = list(axes_dict.values())
    axes_str = ['', '', '']
    n = 0
    for val in axes:
        index = axes_vals.index(val)
        axes_str[n] = axes_keys[index]
        n += 1

    # Create output dictionaries to plot figures
    x_label = "%s axis" % axes_str[1]
    y_label = "%s axis" % axes_str[0]
    title = "Image"
    if axes[2] == 2:  # Sagittal
        title = "Sagittal"
        if axes[0] == 0 and axes[1] == 1:
            image = np.flip(image, axis=0)
            x_label = "(-Y) A | PHASE | P (+Y)"
            y_label = "(-X) I | READOUT | S (+X)"
        else:
            image = np.transpose(image, (0, 2, 1))
            image = np.flip(image, axis=0)
            x_label = "(-Y) A | READOUT | P (+Y)"
            y_label = "(-X) I | PHASE | S (+X)"
        image_orientation_dicom = [0.0, 1.0, 0.0, 0.0, 0.0, -1.0]
    elif axes[2] == 1:  # Coronal
        title = "Coronal"
        if axes[0] == 0 and axes[1] == 2:
            x_label = "(+Z) R | PHASE | L (-Z)"
            y_label = "(-X) I | READOUT | S (+X)"
        else:
            image = np.transpose(image, (0, 2, 1))
            x_label = "(+Z) R | READOUT | L (-Z)"
            y_label = "(-X) I | PHASE | S (+X)"
        image_orientation_dicom = [1.0, 0.0, 0.0, 0.0, 0.0, -1.0]
    elif axes[2] == 0:  # Transversal
        title = "Transversal"
        if axes[0] == 1 and axes[1] == 2:
            image = np.flip(image, axis=0)
            x_label = "(+Z) R | PHASE | L (-Z)"
            y_label = "(+Y) P | READOUT | A (-Y)"
        else:
            image = np.transpose(image, (0, 2, 1))
            image = np.flip(image, axis=0)
            x_label = "(+Z) R | READOUT | L (-Z)"
            y_label = "(+Y) P | PHASE | A (-Y)"
        image_orientation_dicom = [1.0, 0.0, 0.0, 0.0, 1.0, 0.0]

    output = {
        'widget': 'image',
        'data': image,
        'xLabel': x_label,
        'yLabel': y_label,
        'title': title,
    }

    return output, image, image_orientation_dicom

def save_dicom(axes_orientation, n_points, fov, image, file_path, meta_data={}):
    image = np.abs(image)
    axes_orientation = np.array(axes_orientation)
    n_xyz = [0, 0, 0]
    reorder = [0, 0, 0]

    # Determine correct dimension ordering
    for i, axis in enumerate(axes_orientation):
        n_xyz[axis] = n_points[i]
        reorder[axis] = i

    fov = np.array(fov)  # cm
    fov = fov[axes_orientation]
    resolution = fov / n_points * 10  # Convert cm to mm

    # DICOM TAGS
    if (axes_orientation == [0, 1, 2]).all():
        meta_data["ImageOrientationPatient"] = [0.0, 0.0, -1.0, 0.0, 1.0, 0.0]
        meta_data['PixelSpacing'] = [resolution[1], resolution[0]]
        meta_data['SliceThickness'] = resolution[2]
    elif (axes_orientation == [1, 0, 2]).all():
        meta_data["ImageOrientationPatient"] = [0.0, 1.0, 0.0, 0.0, 0.0, -1.0]
        meta_data['PixelSpacing'] = [resolution[1], resolution[0]]
        meta_data['SliceThickness'] = resolution[2]
    elif (axes_orientation == [1, 2, 0]).all():
        image = image[::-1, : ,:]
        meta_data["ImageOrientationPatient"] = [0.0, 1.0, 0.0, 1.0, 0.0, 0.0]
        meta_data['PixelSpacing'] = [resolution[1], resolution[0]]
        meta_data['SliceThickness'] = resolution[2]
    elif (axes_orientation == [2, 1, 0]).all():
        image = image[::-1, : ,:]
        meta_data["ImageOrientationPatient"] = [1.0, 0.0, 0.0, 0.0, 1.0, 0.0]
        meta_data['PixelSpacing'] = [resolution[1], resolution[0]]
        meta_data['SliceThickness'] = resolution[2]
    elif (axes_orientation == [2, 0, 1]).all():
        meta_data["ImageOrientationPatient"] = [1.0, 0.0, 0.0, 0.0, 0.0, -1.0]
        meta_data['PixelSpacing'] = [resolution[1], resolution[0]]
        meta_data['SliceThickness'] = resolution[2]
    elif (axes_orientation == [0, 2, 1]).all():
        meta_data["ImageOrientationPatient"] = [0.0, 0.0, -1.0, 1.0, 0.0, 0.0]
        meta_data['PixelSpacing'] = [resolution[1], resolution[0]]
        meta_data['SliceThickness'] = resolution[2]

    # Normalize to 16-bit (0 - 65535)
    image = (image - np.min(image)) / (np.max(image) - np.min(image)) * (2 ** 15 - 1)
    image = image.astype(np.uint16)

    # Create metadata
    if len(image.shape) > 2:
        # Obtener dimensiones
        slices, rows, columns = image.shape
        meta_data["Columns"] = columns
        meta_data["Rows"] = rows
        meta_data["NumberOfSlices"] = slices
        meta_data["NumberOfFrames"] = slices
    # if it is a 2d image
    else:
        # Obtener dimensiones
        rows, columns = image.shape
        slices = 1
        meta_data["Columns"] = columns
        meta_data["Rows"] = rows
        meta_data["NumberOfSlices"] = 1
        meta_data["NumberOfFrames"] = 1

    meta_data["PixelData"] = image.tobytes()
    meta_data["WindowWidth"] = 26373
    meta_data["WindowCenter"] = 13194

    # Create DICOM object
    dicom_image = DICOMImage()

    # Save image into DICOM object
    dicom_image.meta_data["PixelData"] = meta_data["PixelData"]

    # Date and time
    current_time = datetime.now()
    meta_data["StudyDate"] = current_time.strftime("%Y%m%d")
    meta_data["StudyTime"] = current_time.strftime("%H%M%S")

    # Update the DICOM metadata
    dicom_image.meta_data.update(meta_data)

    # Save metadata dictionary into DICOM object metadata (Standard DICOM 3.0)
    dicom_image.image2Dicom()

    # Save DICOM file
    dicom_image.save(f"{file_path}")
    
    # Save the DICOM file
    print(f"DICOM saved: {file_path}")


def save_nifti(axes_orientation, n_points, fov, dfov, image, file_path):
    axes_orientation = np.array(axes_orientation)
    n_xyz = [0, 0, 0]
    reorder = [0, 0, 0]
    for i, axis in enumerate(axes_orientation):
        n_xyz[axis] = n_points[i]
        reorder[axis] = i
    fov = np.array(fov)  # cm
    dfov = np.array(dfov)  # mm
    resolution = fov / n_xyz * 10  # mm
    image = np.transpose(image, axes=(2, 1, 0))
    image = np.transpose(image, reorder)
    image = np.transpose(image, axes=(2, 1, 0))
    image = image[::-1, ::-1, ::-1]
    image = image / np.max(image)

    orientation = 'FFS'

    # Create afine matrix
    if orientation == 'FFS':
        affine = np.zeros((4, 4))
        affine[0, 0] = resolution[2]
        affine[1, 1] = resolution[1]
        affine[2, 2] = resolution[0]
        affine[0, 3] = dfov[2]
        affine[1, 3] = - dfov[1]
        affine[2, 3] = dfov[0]
        affine[3, 3] = 1
    elif orientation == 'HFS':
        print("Affine matrix not ready.")
    elif orientation == 'FFP':
        print("Affine matrix not ready.")
    elif orientation == 'HFP':
        print("Affine matrix not ready.")

    # Create the NIfTI image
    nifti_img = nib.Nifti1Image(np.abs(image), affine)

    # Save the NIfTI file
    nib.save(nifti_img, file_path)

def run_cosbell_filter(sampled, data, cosbell_order):
    """
    Apply the Cosbell filter operation to the k-space data along three directions.

    This method applies the Cosbell filter to the k-space data along the readout ('rd'), phase ('ph'), and slice
    ('sl') directions. It modifies the input k-space data in-place.

    Args:
        sampled (ndarray): The sampled k-space coordinates in a nx3 matrix, where n is the number of points in k-space. The three columns correspond to the readout, phase, and slice directions.
        data (ndarray): The 3D matrix representing the k-space data to be filtered (sl, ph, rd).
        cosbell_order (list): The order of the Cosbell filter for each direction (rd, ph, sl)

    Returns:
        ndarray: The filtered k-space data (sl, ph, rd).
    """
    n_points = data.shape

    # Along readout
    k = np.reshape(sampled[:, 0], n_points)
    kmax = np.max(np.abs(k[:]))
    theta = k / kmax
    data *= (np.cos(theta * (np.pi / 2)) ** cosbell_order[0])

    # Along phase
    k = np.reshape(sampled[:, 1], n_points)
    kmax = np.max(np.abs(k[:]))
    theta = k / kmax
    data *= (np.cos(theta * (np.pi / 2)) ** cosbell_order[1])

    # Along slice
    k = np.reshape(sampled[:, 2], n_points)
    kmax = np.max(np.abs(k[:]))
    theta = k / kmax
    data *= (np.cos(theta * (np.pi / 2)) ** cosbell_order[2])

    return data

def run_bm4d_filter(image_data, std=0):
    """
    Apply the BM4D filter to denoise the image.

    This method retrieves the image data, rescales it, calculates the standard deviation for the BM4D filter,
    applies the BM4D filter to denoise the rescaled image, and rescales the denoised image back to its original
    scale.

    Args:
        image_data (ndarray): The input image data.
        std (float): standard deviation for bm4d. If zero, the standard deviation will be calculated automatically. Default 0

    Returns:
        ndarray: The denoised image.

    """
    # Rescale the image data for processing
    reference = np.max(image_data)
    image_rescaled = image_data / reference * 100

    # Calculate the standard deviation for BM4D filter
    if std==0:
        # Quantize image
        num_bins = 1000
        image_quantized = np.digitize(image_rescaled, bins=np.linspace(0, 1, num_bins + 1)) - 1

        # Divide the image into blocks
        n_multi = (np.array(image_quantized.shape) / 5).astype(int) * 5
        blocks_r = view_as_blocks(image_rescaled[0:n_multi[0], 0:n_multi[1], 0:n_multi[2]], block_shape=(5, 5, 5))

        # Calculate the standard deviation for each block
        block_std_devs = np.std(blocks_r, axis=(3, 4, 5))

        # Calculate the average value for each block
        block_mean = np.mean(blocks_r, axis=(3, 4, 5))

        # Find the indices of the block with the minimum mean
        min_mean_index = np.unravel_index(np.argmin(block_mean), block_mean.shape)

        # Extract the block with the highest entropy from the block_std_devs array
        std = 4 * block_std_devs[min_mean_index]
        print("Standard deviation for BM4D: %0.2f" % std)

    # Create a BM4D profile and set options
    profile = bm4d.BM4DProfile()
    stage_arg = bm4d.BM4DStages.ALL_STAGES
    blockmatches = (False, False)

    # Apply the BM4D filter to the rescaled image
    denoised_rescaled = bm4d.bm4d(image_rescaled, sigma_psd=5, profile=profile, stage_arg=stage_arg,
                                  blockmatches=blockmatches)

    # Rescale the denoised image back to its original dimensions
    denoised_image = denoised_rescaled / 100 * reference

    return denoised_image

def run_dfft(image):
    """
    Perform direct FFT reconstruction.

    This method performs the direct Fourier transform to obtain the k-space data from an image in the spatial
    domain.

    Args:
        image (ndarray): The image in the spatial domain (sl, ph, rd).

    Returns:
        ndarray: The k-space data.

    """
    k_space = np.fft.fftshift(np.fft.fftn(np.fft.ifftshift(image)))
    return k_space

def run_ifft(k_space):
    """
    Perform inverse FFT reconstruction.

    This method performs the inverse Fourier transform to reconstruct the image in the spatial domain from k-space
    data.

    Args:
        k_space (ndarray): The k-space data (sl, ph, rd).

    Returns:
        ndarray: The reconstructed image in the spatial domain.

    """
    image = np.fft.ifftshift(np.fft.ifftn(np.fft.fftshift(k_space)))
    return image

def run_zero_padding(k_space, new_size):
    """
    Perform zero-padding on the given k-space data.

    This function retrieves the desired zero-padding dimensions from the user input,
    creates a new k-space matrix with the specified size filled with zeros, and places
    the original k-space data at the center of the new matrix. It ensures that the
    k-space is properly centered even if the new dimensions are larger or smaller
    than the original.

    Parameters:
    -----------
    k_space : np.ndarray
        A 3D complex-valued NumPy array representing the original k-space data.
    new_size: list
        A list with the new matrix size (rd, ph, sli)

    Returns:
    --------
    np.ndarray
        A 3D complex-valued NumPy array containing the new k-space data.

    Notes:
    ------
    - The function calculates the appropriate offsets to center the original k-space
      within the new zero-padded matrix.
    - If the new matrix size is smaller in any dimension, it crops the k-space accordingly.
    """
    # Get the k_space shape
    shape_0 = k_space.shape

    # Determine the new shape after zero-padding
    n_rd = int(new_size[0])
    n_ph = int(new_size[1])
    n_sl = int(new_size[2])
    shape_1 = n_sl, n_ph, n_rd

    # Create an image matrix filled with zeros
    image_matrix = np.zeros(shape_1, dtype=complex)

    # Calculate the centering offsets for each dimension
    offset_0 = (shape_1[0] - shape_0[0]) // 2
    offset_1 = (shape_1[1] - shape_0[1]) // 2
    offset_2 = (shape_1[2] - shape_0[2]) // 2

    # Calculate the start and end indices to center the k_space within the new image_matrix
    new_start_0 = offset_0 if offset_0 >= 0 else 0
    new_start_1 = offset_1 if offset_1 >= 0 else 0
    new_start_2 = offset_2 if offset_2 >= 0 else 0
    new_end_0 = new_start_0 + shape_0[0] if offset_0 > 0 else shape_1[0]
    new_end_1 = new_start_1 + shape_0[1] if offset_1 > 0 else shape_1[1]
    new_end_2 = new_start_2 + shape_0[2] if offset_2 > 0 else shape_1[2]

    # Calculate the start and end indices of old matrix
    old_start_0 = 0 if offset_0 >= 0 else -offset_0
    old_start_1 = 0 if offset_1 >= 0 else -offset_1
    old_start_2 = 0 if offset_2 >= 0 else -offset_2
    old_end_0 = shape_0[0] if offset_0 >= 0 else old_start_0 + shape_1[0]
    old_end_1 = shape_0[1] if offset_1 >= 0 else old_start_1 + shape_1[1]
    old_end_2 = shape_0[2] if offset_2 >= 0 else old_start_2 + shape_1[2]

    # Copy the k_space into the image_matrix at the center
    image_matrix[new_start_0:new_end_0, new_start_1:new_end_1, new_start_2:new_end_2] = k_space[
                                                                                        old_start_0:old_end_0,
                                                                                        old_start_1:old_end_1,
                                                                                        old_start_2:old_end_2
                                                                                        ]

    return image_matrix

def hanning_filter(kSpace, mm, nb_point):
    """
    Apply a Hanning filter to the k-space data.

    Args:
        kSpace (ndarray): K-space data.
        n (int): Number of zero-filled points in k-space.
        m (int): Number of acquired points in k-space.
        nb_point (int): Number of points before m+n where reconstruction begins to go to zero.

    Returns:
        ndarray: K-space data with the Hanning filter applied.
    """
    kSpace_hanning = np.copy(kSpace)
    kSpace_hanning[mm[0]::, :, :] = 0.0
    kSpace_hanning[:, mm[1]::, :] = 0.0
    kSpace_hanning[:, :, mm[2]::] = 0.0

    # Calculate the Hanning window
    hanning_window = np.hanning(nb_point * 2)
    hanning_window = hanning_window[int(len(hanning_window)/2)::]

    if not mm[0] == np.size(kSpace, 0):
        for ii in range(nb_point):
            kSpace_hanning[mm[0]-nb_point+ii+1, :, :] *= hanning_window[ii]
    if not mm[1] == np.size(kSpace, 1):
        for ii in range(nb_point):
            kSpace_hanning[:, mm[1]-nb_point+ii+1, :] *= hanning_window[ii]
    if not mm[2] == np.size(kSpace, 2):
        for ii in range(nb_point):
            kSpace_hanning[:, :, mm[2]-nb_point+ii+1] *= hanning_window[ii]

    return kSpace_hanning

def run_pocs_reconstruction(n_points, factors, k_space_ref):
    """
    Perform POCS reconstruction.

    Retrieves the number of points before m+n where reconstruction begins to go to zero.
    Retrieves the correlation threshold for stopping the iterations.
    Computes the partial image and full image for POCS reconstruction.
    Applies the iterative reconstruction with phase correction.
    Updates the main matrix of the image view widget with the interpolated image.
    Adds the "POCS" operation to the history widget and updates the history dictionary and operations history.
    """

    def getCenterKSpace(k_space, m_vec):
        # fix n_vec
        output = np.zeros(np.shape(k_space), dtype=complex)
        n_vec = np.array(np.shape(k_space))

        # fill with zeros
        idx0 = n_vec // 2 - m_vec
        idx1 = n_vec // 2 + m_vec
        output[idx0[0]:idx1[0], idx0[1]:idx1[1], idx0[2]:idx1[2]] = \
            k_space[idx0[0]:idx1[0], idx0[1]:idx1[1], idx0[2]:idx1[2]]

        return output

    # Get n and m
    factors = [float(num) for num in factors][-1::-1]
    mm = np.array([int(num) for num in (n_points * factors)])
    m = np.array([int(num) for num in (n_points * factors - n_points / 2)])

    # Get the reference image
    img_ref = np.abs(run_ifft(k_space_ref))

    # Create a copy with the center of k-space
    k_space_center = getCenterKSpace(k_space_ref, m)

    # Number of points before m+n where we begin to go to zero
    nb_point = 2

    # Set the correlation threshold for stopping the iterations
    threshold = 1e-6

    # Get image phase
    img_center = run_ifft(k_space_center)
    phase = img_center / abs(img_center)

    # Generate the corresponding image with the Hanning filter
    k_space_hanning = hanning_filter(k_space_ref, mm, nb_point)
    img_hanning = np.abs(run_ifft(k_space_hanning))

    num_iterations = 0  # Initialize the iteration counter
    previous_img = img_hanning.copy()  # you have the choice between img_hanning or img_ramp

    while True:
        # Iterative reconstruction
        img_iterative = previous_img * phase
        k_space_new = run_dfft(img_iterative)

        # Apply constraint: Keep the region of k-space from n+m onwards and restore the rest
        k_space_new[0:mm[0], 0:mm[1], 0:mm[2]] = k_space_ref[0:mm[0], 0:mm[1], 0:mm[2]]

        # Reconstruct the image from the modified k-space
        img_reconstructed = np.abs(run_ifft(k_space_new))

        # Compute correlation between consecutive reconstructed images
        correlation = np.corrcoef(previous_img.flatten(), img_reconstructed.flatten())[0, 1]

        # Display correlation and current iteration number
        print("Iteration: %i, Convergence: %0.2e" % (num_iterations, (1-correlation)))

        # Check if correlation reaches the desired threshold
        if (1-correlation) <= threshold or num_iterations >= 100:
            break

        # Update previous_img for the next iteration
        previous_img = img_reconstructed.copy()

        # Increment the iteration counter
        num_iterations += 1

    # Get correlation with reference image
    correlation = np.corrcoef(img_ref.flatten(), img_reconstructed.flatten())[0, 1]
    print("Respect the reference image:")
    print("Convergence: %0.2e" % (1 - correlation))

    return img_reconstructed

if __name__ == "__main__":
    # mat_data = sp.io.loadmat("/home/physio/git_repos/Results/Dicom/RarePyPulseq.2025.03.27.07.10.03.769.mat")
    # mat_data = sp.io.loadmat("/home/physio/git_repos/Results/Dicom/RarePyPulseq.2025.03.27.07.10.44.360.mat")
    # mat_data = sp.io.loadmat("/home/physio/git_repos/Results/Dicom/RarePyPulseq.2025.03.27.07.11.22.935.mat")
    # mat_data = sp.io.loadmat("/home/physio/git_repos/Results/Dicom/RarePyPulseq.2025.03.27.07.12.09.496.mat")
    # mat_data = sp.io.loadmat("/home/physio/git_repos/Results/Dicom/RarePyPulseq.2025.03.27.07.12.52.711.mat")
    mat_data = sp.io.loadmat("/home/physio/git_repos/Results/Dicom/RarePyPulseq.2025.03.27.07.13.29.064.mat")
    image = np.abs(mat_data['image3D'])
    fov = mat_data['fov'][0]
    n_points = mat_data['nPoints'][0]
    dfov = mat_data['dfov'][0]
    axes_orientation = mat_data['axesOrientation'][0]
    file_path = "/home/physio/git_repos/Results/Dicom/new_dicom.dcm"
    save_dicom(axes_orientation, n_points, fov, image, file_path)

    # image = np.zeros((10, 12, 14))
    # image[0:3, 0, 0] = 1
    # image[0, 0:5, 0] = 1
    # image[0, 0, 0:7] = 1
    # fov = np.array([10, 12, 14])
    # n_points = np.array([14, 12, 10])
    # dfov = np.array([0.0, 0.0, 0.0])
    # axes_orientation = np.array([0, 1, 2])
    # file_path = "/home/physio/git_repos/Results/Dicom/new_dicom_1.dcm"
    # save_dicom(axes_orientation, n_points, fov, dfov, image, file_path)




# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys
sys.path.insert(0, os.path.abspath('../../..'))

project = 'MaRGE'
copyright = '2024, J.M. Algarín'
author = 'J.M. Algarín'
release = '0.8.0'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.doctest',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode'
]

templates_path = ['_templates']
exclude_patterns = []

autodoc_default_options = {
    'members': True,
    'undoc-members': False,
    'show-inheritance': False
}

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
# html_theme = 'neutral'
html_static_path = ['_static']
add_module_names = False



#!/usr/bin/env python3
# Basic CSV -> machine code compiler for marga

import numpy as np
import warnings
from marmachine import *
try:
    from local_config import grad_board
except ModuleNotFoundError:
    grad_board = "gpa-fhdo"

grad_data_bufs = (1, 2)

max_removed_instructions = 1000

def debug_print(*args, **kwargs):
    # print(*args, **kwargs)
    pass

def col2buf(col_idx, value):
    """ Returns a tuple of (buffer indices), (values), (value masks)
    Value masks specify which bits are actually relevant on the output.
    Can accept arrays of values."""
    if col_idx in (1, 2, 3, 4): # TX
        buf_idx = col_idx + 4, # TX0_I, TX0_Q, TX1_I, TX1_Q
        val = value,
        mask = 0xffff,
    elif col_idx in (5, 6, 7, 8, 9, 10, 11, 12): # grad
        # Only encode value and channel into words here.  Precise
        # timing and broadcast logic will be handled at the next stage
        if grad_board == "gpa-fhdo":
            if col_idx in (9, 10, 11, 12):
                raise RuntimeError("GPA-FHDO is selected, but you are trying to control OCRA1")
            grad_chan = col_idx - 5
            val_full = value | 0x80000 | ( grad_chan << 16 ) | (grad_chan << 25)
        elif grad_board == "ocra1":
            if col_idx in (5, 6, 7, 8):
                raise RuntimeError("OCRA1 is selected, but you are trying to control GPA-FHDO")
            grad_chan = col_idx - 9
            val_full = value << 2 | 0x00100000 | (grad_chan << 25) | 0x01000000 # always broadcast by default
        else:
            raise ValueError("Unknown grad board")

        buf_idx = 2, 1 # GRAD_MSB, GRAD_LSB
        val = val_full >> 16, val_full & 0xffff
        mask = 0xffff, 0xffff
    elif col_idx in (13, 14): # RX rate
        buf_idx = col_idx - 10, # RX0_RATE, RX1_RATE
        val = value,
        mask = 0xffff,
    elif col_idx in (15, 16): # RX rate valid
        buf_idx = 16, # RX_CTRL
        bit_idx = col_idx - 15
        val = value << (4 + bit_idx),
        mask = 0x1 << (4 + bit_idx),
    elif col_idx in (17, 18): # RX resets, active low
        buf_idx = 16, # RX_CTRL
        bit_idx = col_idx - 17
        val = value << (6 + bit_idx),
        mask = 0x1 << (6 + bit_idx),
    elif col_idx in (19, 20): # RX resets, active low
        buf_idx = 16, # RX_CTRL
        bit_idx = col_idx - 19
        val = value << (8 + bit_idx),
        mask = 0x1 << (8 + bit_idx),
    elif col_idx in (21, 22, 23): # TX/RX gates, external trig
        buf_idx = 15, # GATES_LEDS
        bit_idx = col_idx - 21
        val = value << bit_idx,
        mask = 0x1 << bit_idx,
    elif col_idx == 24: # LEDs
        buf_idx = 15, # GATES_LEDS
        val = value << 8,
        mask = 0xff00,
    elif col_idx in (25, 26, 27): # LO freqs
        lo_lsb_buf = 9 + 2*(col_idx - 25) # 9, 11 or 13
        buf_idx = lo_lsb_buf, lo_lsb_buf + 1 # DDS[0,1,2]_PHASE_LSB, DDS[0,1,2]_PHASE_MSB
        val = value & 0xffff, value >> 16
        mask = 0xffff, 0x7fff
    elif col_idx in (28, 29, 30): # LO phase reset
        lo_msb_buf = 10 + 2*(col_idx - 28) # DDS[0,1,2]_PHASE_MSB
        buf_idx = lo_msb_buf,
        val = value << 15,
        mask = 0x8000,
    elif col_idx in (31, 32): # LO source for RX demodulation
        buf_idx = 16, # RX_CTRL
        bit_idx = (col_idx - 31) * 2
        val = value << bit_idx,
        mask = 0x0003 << bit_idx,

    return np.uint16(buf_idx), np.uint16(val), np.uint16(mask)

def csv2bin(path, quick_start=False, initial_bufs=np.zeros(MARGA_BUFS, dtype=np.uint16), latencies = np.zeros(MARGA_BUFS, dtype=np.int32)):
    """ initial_bufs: starting state of output buffers, to track with instructions
    quick_start: strip out the initial RAM-writing dead time if the CSV was generated by the simulator or similar
    latencies: inherent buffer latencies to take into
    account. Latencies are primarily relevant to the gradients, but
    can be adjusted to suit various other external hardware effects
    like slow RF amps, very long cables etc
    """

    # Input: CSV column, starting from 0 for tx0 i and ending with 21 for leds
    # Output: corresponding buffer index or indices to change

    data = np.loadtxt(path, skiprows=1, delimiter=',', comments='#').astype(np.uint32)
    with open(path, 'r') as csvf:
        cols = csvf.readline().strip().split(',')[1:]

    assert cols[-1] == ' csv_version_0.2', "Wrong CSV format"

    if quick_start:
        # remove dead time in the beginning taken up by simulated memory writes, if the input CSV is generated from the simulator
        # data[1:, 0] = data[1:, 0] - data[1, 0] + latencies.max()
        data[1:, 0] = data[1:, 0] - data[1, 0] + 10

    # Boolean: compare data offset by one row in time
    data_diff = data[:-1,1:] != data[1:,1:]

    changelist = []
    changelist_grad = []

    for k, dd in enumerate(data_diff):
        clocktime = data[k + 1, 0]
        dw = np.where(dd)[0] # indices where data changed
        for col_idx, value in zip(dw + 1, data[k + 1][dw + 1]):
            buf_idces, vals, masks = col2buf(col_idx, value)
            for bi, v, m in zip(buf_idces, vals, masks):
                change = clocktime - latencies[bi], bi, v, m
                if bi in grad_data_bufs:
                    changelist_grad.append(change)
                else:
                    changelist.append(change)

    return cl2bin(changelist, changelist_grad, initial_bufs)

def dict2bin(sd, initial_bufs=np.zeros(MARGA_BUFS, dtype=np.uint16), latencies = np.zeros(MARGA_BUFS, dtype=np.int32)):
    """sd: sequence dictionary, consisting of something in the form of:

     {'tx0_i': ( np.array([100, 102, 304, 506]), np.array([1, 200, 65535, 20000]) ),
      'fhdo_vx': ( np.array([3000, 4500, 5900, 7000]), np.array([1, 2, 55555, 33333]) ),
      'fhdo_vy': ( np.array([10000, 12000, 14000, 16000]), np.array([1, 2, 55555, 33333]) ) }

    etc. Same binary format as in the CSV file.

    latencies: inherent buffer latencies to take into
    account. Latencies are primarily relevant to the gradients, but
    can be adjusted to suit various other external hardware effects
    like slow RF amps, very long cables etc
    """

    col_arr = ['clock cycles', 'tx0_i', 'tx0_q', 'tx1_i', 'tx1_q', 'fhdo_vx', 'fhdo_vy', 'fhdo_vz', 'fhdo_vz2',
               'ocra1_vx', 'ocra1_vy', 'ocra1_vz', 'ocra1_vz2', 'rx0_rate', 'rx1_rate',
               'rx0_rate_valid', 'rx1_rate_valid', 'rx0_rst_n', 'rx1_rst_n', 'rx0_en', 'rx1_en',
               'tx_gate', 'rx_gate', 'trig_out', 'leds',
               'lo0_freq', 'lo1_freq', 'lo2_freq', 'lo0_rst', 'lo1_rst', 'lo2_rst',
               'rx0_lo', 'rx1_lo', ] # TODO: these two rows aren't yet in the CSV and thus aren't tested by test_marga_model.py

    changelist = []
    changelist_grad = []

    for k, vals in sd.items(): # iterate over dictionary keys
        col_idx = col_arr.index(k)
        changelist_grad_local = []
        buf_idces, values, masks = col2buf(col_idx, vals[1]) # single element or array of values
        t_corr = vals[0] - latencies[buf_idces[0]]
        for bi, vv, m in zip(buf_idces, values, masks):
            for t, v in zip(t_corr, vv):
                change = t, bi, v, m
                if bi in grad_data_bufs:
                    changelist_grad_local.append(change)
                else:
                    changelist.append(change)

        # needed to keep coupled LSB/MSB pairs together in case
        # multiple events occur on different channels simultaneously
        if len(changelist_grad_local) != 0:
            changelist_grad_local.sort(key=lambda change: change[0])
            changelist_grad += changelist_grad_local

    return cl2bin(changelist, changelist_grad, initial_bufs)

def cl2bin(changelist, changelist_grad,
           initial_bufs=np.zeros(MARGA_BUFS, dtype=np.uint16)):

    """Central compilation function; accept in two changelists,
    changelist for all the direct-buffer outputs (TX, most configurable
    parameters, etc) and the other, changelist_grad, for the outputs used
    to control hardware with non-trivial internal timing behaviour
    (currently only the gradient boards). Also accepts non-default initial
    values to program the buffers to."""

    # Process the grad changelist, depending on what GPA is being used etc
    # Sort in pairs of changes, because otherwise channels can get mixed up
    changelist_grad_paired = [ [k, m] for k, m in zip(changelist_grad[::2], changelist_grad[1::2]) ]
    sortfn = lambda change: change[0]
    # changelist_grad.sort(key=sortfn) # sort by time
    sortfn_paired = lambda change: change[0][0]
    changelist_grad_paired.sort(key=sortfn_paired) # sort by time
    changelist_grad = [k for sl in changelist_grad_paired for k in sl] # https://stackabuse.com/python-how-to-flatten-list-of-lists/

    t_last = [0, 0] # no updates have previously happened; [LSB, MSB]
    spi_div = (initial_bufs[0] & 0xfc) >> 2
    changelist_grad_shifted = []
    num_chgs = [0, 0] # [LSB, MSB]
    grad_vals = [initial_bufs[1], initial_bufs[2]] # [LSB, MSB] current output data
    grad_vals_old = [0, 0] # [LSB, MSB] previous output data

    for c in changelist_grad:
        t = c[0]
        debug_print("t: ", t, " t_last: ", t_last, "num_chgs: ", num_chgs, " c: ", c)
        idx = c[1] - 1 # 0 for LSB, 1 for MSB
        msb = idx == 1
        data = c[2]
        # if data == grad_vals[idx]: # no actual change to buffer output
        #     continue # skip this change
        # else:
        #     grad_vals_old[idx] = grad_vals[idx]
        #     grad_vals[idx] = data # update the last known buffer value

        if t == t_last[idx]:
            num_chgs[idx] += 1
            # assume the changes in changelist_grad are paired with LSBs/MSBs matching each other's grad channels stored sequentially,
            # and that for each event, the MSB update is first
            if grad_board == "ocra1": # simultaneous with another grad update
                if msb:
                    if num_chgs[1]: # MSB buffer and not the first grad event on this timestep
                        # turn broadcast off if this isn't the first grad event on this timestep
                        data = data & ~np.uint16(0x0100)
                        # return LSB back to old values, since this one is now done in the past
                        grad_vals[:] = grad_vals_old # revert the last known buffer values
                # else:
                #     if data == grad_vals[idx]: # no actual change to buffer output compared to earlier LSB at this timestep
                #         continue # skip this change

                # move non-broadcast events back in time, so that synchronisation will be done in ocra1_iface core
                changelist_grad_shifted.append( (c[0]-num_chgs[idx], c[1], data, c[3]) )
                num_chgs[idx] += 1
            elif grad_board == "gpa-fhdo":
                # don't do anything; currently will cause an error
                # later since multiple events can't happen at the same
                # time for GPA-FHDO
                changelist_grad_shifted.append(c)
        else:
            if t - t_last[idx] < 24 * (1 + spi_div) + 2: #
                warnings.warn("Gradient updates are too frequent for selected SPI divider. Missed samples are likely!", MarGradWarning)

            # if data == grad_vals[idx]: # no actual change to buffer output
            #     continue # skip this change

            t_last[idx] = t
            grad_vals[idx] = data # update the last known buffer value
            changelist_grad_shifted.append(c)
            num_chgs = [0, 0]

    changelist += changelist_grad_shifted
    changelist.sort(key=sortfn) # sort by time

    # Track removed instruction events, but only warn when the number exceeds a minimum
    removed_instruction_warnings = []

    # Process and combine the change list into discrete sets of operations at each time, i.e. an output list
    def cl2ol(changelist):
        current_bufs = initial_bufs.copy()
        current_time = changelist[0][0]
        unique_times = []
        unique_changes = []
        change_masks = np.zeros(MARGA_BUFS, dtype=np.uint16)
        changed = np.zeros(MARGA_BUFS, dtype=bool)

        def close_timestep(time):
            ch_idces = np.where(changed)[0]
            # buf_time_offsets = np.zeros(MARGA_BUFS, dtype=int32)
            buf_time_offsets = 0
            unique_changes.append( [time, ch_idces, current_bufs[ch_idces], buf_time_offsets] )
            change_masks[:] = np.zeros(MARGA_BUFS, dtype=np.uint16)
            changed[:] = np.zeros(MARGA_BUFS, dtype=bool)

        for time, buf, val, mask in changelist:
            if time != current_time:
                close_timestep(current_time)
                current_time = time
            buf_diff = (current_bufs[buf] ^ val) & mask
            assert buf_diff & change_masks[buf] == 0, "Tried to set a buffer to two values at once"
            if buf_diff == 0:
                if buf not in (1, 2):
                    # gradient buffers will have unneeded instructions
                    # all the time, so not worth warning the user for
                    # those
                    removed_instruction_warnings.append( "Instruction at tick {:d}, buffer {:d}, value 0x{:04x}, mask 0x{:04x} will have no effect. Skipping...".format(time, buf, val, mask) )
                continue
            val_masked = val & mask
            old_val_unmasked = current_bufs[buf] & ~mask
            new_val = old_val_unmasked | val_masked
            change_masks[buf] |= mask
            current_bufs[buf] = new_val
            changed[buf] = True

        close_timestep(current_time)

        return unique_changes

    changes = cl2ol(changelist)

    # warn about all the removed instructions if there are more than a maximum number
    if len(removed_instruction_warnings) > max_removed_instructions:
        for riw in removed_instruction_warnings:
            warnings.warn(riw, MarRemovedInstructionWarning)
        warnings.warn("NOTE: Fewer than {:d} removed-instruction warnings will not be printed -- keep this in mind when searching for the root cause.".format(max_removed_instructions))

    # Process time offsets
    for ch, ch_prev in zip( reversed(changes[1:]), reversed(changes[:-1]) ):
        # does the current timestep need to output more data than can
        # fit into the time gap since the previous timestep?
        timestep = np.int32(ch[0] - ch_prev[0])
        timediff = np.int32(ch[1].size - timestep)
        # if timestep < ch[1].size: # not enough time

        if timediff > 0:
            ch_prev[0] -= timediff # move prev. event into the past
            ch_prev[3] = timediff # make prev. event's buffers output in its future

    # convert to differential timesteps
    last_time = 0
    for ch in changes:
        ch0 = ch[0]
        ch[0] = ch0 - last_time
        last_time = ch0

    # Interpretation of each element of changes list:
    # [time when all instructions for this change will have completed,
    #  buffers that need to be changed,
    #  values to set the buffers to,
    #  the delay until the buffers will output their values]

    ### Write out instructions

    # Write out initial buffer values
    bdata = []
    addr = 0
    states = initial_bufs
    # reversed order, so that grad board is enabled last of all (to avoid spurious initial transfer)
    for k, ib in enumerate(reversed(initial_bufs)):
        bdata.append(instb(MARGA_BUFS-1-k, k, ib))

    last_buf_time_left = np.zeros(MARGA_BUFS, dtype=np.int32)
    buf_time_left = np.zeros(MARGA_BUFS, dtype=np.int32)
    # buf_empty_time = np.zeros(MARGA_BUFS, dtype=np.int32)
    debug_print("changes:")
    for k in changes:
        debug_print(k)

    for event in changes:
        b_instrs = event[1].size
        dtime = event[0]

        # soak up any extra time which is in excess of what the instructions need to execute synchronously
        excess_dtime = dtime - b_instrs
        excess_dtime_tmp = excess_dtime
        while excess_dtime_tmp > 2: # delay of 3 or more cycles needed
            wait_time = min(excess_dtime_tmp, COUNTER_MAX + 3) # delay for the time instruction
            bdata.append(insta(IWAIT, wait_time - 3))
            excess_dtime_tmp -= wait_time
            debug_print("i wait ", wait_time - 3)
        if excess_dtime_tmp: # final delay of 1 or 2 cycles
            for k in range(dtime - b_instrs):
                debug_print("i nop")
                bdata.append(insta(INOP, 0))

        # time left after delays from nops or waits
        # dtime_eff could be increased later with a more advanced
        # compiler, to make the buffers bear more of the internal
        # delays
        # dtime_eff = b_instrs

        # count down the times until each channel buffer will be empty
        buf_time_left -= excess_dtime
        buf_time_left[buf_time_left < 0] = 0
        this_time_offset = event[3]
        debug_print("--- dtime {:2d}, this_time_offset: {:2d}, b_instrs: {:2d}, lbtl: ".format(dtime, this_time_offset, b_instrs), last_buf_time_left[5:9])
        for m, (ind, dat) in enumerate(zip(event[1], event[2])):
            execution_delay = b_instrs - m - 1 #+ time - 2
            btli = buf_time_left[ind]
            buf_empty = btli <= m
            if buf_empty: # buffer empty for this instruction; need an appropriate delay only for sync
                # (check against m since with successive cycles, remaining buffers will empty out)
                extra_delay = execution_delay + this_time_offset
                buf_time_left[ind] = this_time_offset + b_instrs
            else:
                # buffer already not empty on this cycle
                extra_delay = this_time_offset - btli + b_instrs - 1
                buf_time_left[ind] += extra_delay + 1

            debug_print("bti={:d} btli={:d} m={:d} empty={:d} edel={:d} instb i {:d} del {:d} dat {:d}".format(
                buf_time_left[ind], btli, m, buf_empty, execution_delay, ind, extra_delay, dat))
            bdata.append(instb(ind, extra_delay, dat))

        buf_time_left -= b_instrs # take into account execution time of this timestep

    # Finish sequence
    bdata.append(insta(IFINISH, 0))
    return bdata

CIC_SLOWEST_RATE_NEAREST_POW2 = 1 << np.ceil(np.log2(CIC_SLOWEST_RATE)).astype(int)

def cic_words(rate, set_cic_shift=False):
    # Calculate the data words to transfer to the CIC for a given rate

    # FLoating-point calculation
    assert np.all( (CIC_FASTEST_RATE <= rate) & (rate <= CIC_SLOWEST_RATE) ), "RX rate outside valid range"
    gain_factor_log2 = CIC_STAGES * np.log2( CIC_SLOWEST_RATE_NEAREST_POW2 / rate )
    gain_shift = np.int32(gain_factor_log2) # rounded down
    a = (1 << CIC_RATE_DATAWIDTH) | gain_shift
    b = (0 << CIC_RATE_DATAWIDTH) | rate
    excess_factor = 2**(gain_factor_log2 - gain_shift)
    # b = (2 << rate_datawidth) | int(factor_excess * (1 << (rate_datawidth - 1)) ) # TODO: tell Benjamin about assumed 1 - i.e. save a bit for multiplicand by assuming it's between 1 and 2

    if set_cic_shift:
        return (a, b), excess_factor
    else:
        return (b,), excess_factor

if __name__ == "__main__":
    csv2bin("/tmp/marga.csv")



"""
@author:    José Miguel Algarín
@email:     josalggui@i3m.upv.es
@affiliation:MRILab, i3M, CSIC, Valencia, Spain
"""
import pyqtgraph as pg


class PlotSmithChartWidget(pg.GraphicsLayoutWidget):
    def __init__(self):
        super(PlotSmithChartWidget, self).__init__()

        # Add label to show data from the cross hair
        self.label2 = pg.LabelItem(justify='left')
        self.addItem(self.label2)

        # Create set of colors
        self.pen = [[255, 0, 0],
                    [0, 255, 0],
                    [255, 255, 0],
                    [0, 0, 255],
                    [255, 0, 255],
                    [0, 255, 255],
                    [255, 255, 255],
                    [128, 128, 128]]

        # Create plot_item to show the figure
        self.plot_item = self.addPlot(row=1, col=0)
        self.plot_item.addLegend()

        # Add lines to show cursor
        self.crosshair_v = pg.InfiniteLine(angle=90, movable=False)
        self.crosshair_h = pg.InfiniteLine(angle=0, movable=False)
        self.plot_item.addItem(self.crosshair_v, ignoreBounds=True)
        self.plot_item.addItem(self.crosshair_h, ignoreBounds=True)
        self.proxy = pg.SignalProxy(self.plot_item.scene().sigMouseMoved, rateLimit=60, slot=self.mouseMoved)



#!/usr/bin/env python3

import numpy as np
import matplotlib.pyplot as plt
import experiment as ex
from local_config import grad_board

import pdb
st = pdb.set_trace

def test_jitter(
        rf_pulse_freq = 2, # MHz
        rf_pulse_start_time = 10, # us
        rf_pulse_length = 2, # us
        rf_amp=0.8, # fraction of full-scale
        rf_interval=5, # us, delay between start of first and start of
                       # second pulse - must be bigger than pulse
                       # length
        grad_pulse_start_time = 10, # us
        grad_pulse_length = 40, # us
        grad_interval=80, # us
        grad_amp=0.2, # fraction of full-scale
        loops=10,
        print_loops=False,
        plot_sequence=False,
        experiment_kwargs={}):
    """Two successive pulses, with an adjustable interval between them.

    Trigger an oscilloscope from the TX gate or first pulse, and look
    at the second pulse, with a suitable hold-off time so that
    triggering is reliable. Over many repetitions, the timing jitter
    will be visible.
    """
    tx_gate_start_time = 0
    tx_gate_length = 0.1

    rf_data = (
        rf_pulse_start_time + np.array([0, rf_pulse_length, rf_interval, rf_pulse_length + rf_interval]),
        np.array([rf_amp, 0, rf_amp, 0])
    )

    grad_data = (
        grad_pulse_start_time + np.array([0, grad_pulse_length, grad_interval, grad_pulse_length + grad_interval]),
        np.array([grad_amp, 0, grad_amp, 0])
    )

    gpa_fhdo_offset_time = 10 if grad_board == "gpa-fhdo" else 0

    event_dict = {'tx0': rf_data,
                  'tx1': rf_data,
                  'grad_vx': grad_data,
                  'grad_vy': grad_data,
                  'grad_vz': grad_data,
                  'grad_vz2': grad_data,
                  'tx_gate': (tx_gate_start_time + np.array([0, tx_gate_length]), np.array([1, 0]))}

    exp = ex.Experiment(lo_freq=rf_pulse_freq, gpa_fhdo_offset_time=gpa_fhdo_offset_time, **experiment_kwargs)
    exp.add_flodict(event_dict)

    if plot_sequence:
        exp.plot_sequence()
        plt.show()

    for k in range(loops):
        exp.run()
        if print_loops and k % 1000 == 0:
            print(f"Loop {k}")

    exp.close_server(only_if_sim=True)

if __name__ == "__main__":
    # slowed down to suit VN's setup, due to GPA-FHDO communication chip being dead
    experiment_kwargs = {'init_gpa': True, 'grad_max_update_rate': 0.1}
    loops = int(1e6)

    # Enable the test that you want to run below

    # RF pulses close together, gradient pulses further apart, rf and grad start simultaneously (default)
    if False:
        test_jitter(loops=loops, experiment_kwargs=experiment_kwargs) # many loops

    # RF and gradient pulses start almost simultaneously (RF offset to avoid SPI noise), second pulses happen at an adjustable interval
    if True:
        interval=100000 # 100ms
        amp = 0.2
        test_jitter(loops=loops, rf_pulse_start_time=10.5,
                    rf_interval=interval, rf_amp=amp,
                    grad_interval=interval, grad_amp=amp,
                    experiment_kwargs=experiment_kwargs) # many loops



from PyQt5.QtCore import QEvent

from marge.ui.window_postprocessing import MainWindow

class ProcessingWindowController(MainWindow):
    def __init__(self, *args, **kwargs):
        super(ProcessingWindowController, self).__init__(*args, **kwargs)

    def set_console(self):
        self.left_layout.addWidget(self.main.console)

    def set_session(self, session):
        self.session = session
        self.setWindowTitle(session['directory'])

    def changeEvent(self, event):
        if event.type() == QEvent.ActivationChange:  # Event type 99
            if self.isActiveWindow():
                self.set_console()
        super().changeEvent(event)



"""
:author:    J.M. Algarín
:email:     josalggui@i3m.upv.es
:affiliation: MRILab, i3M, CSIC, Valencia, Spain
"""

import os
import time
import shutil
import platform
import subprocess
import threading
import numpy as np

from marge.widgets.widget_toolbar_marcos import MarcosToolBar
import marge.marcos.marcos_client.experiment as ex
import marge.configs.hw_config as hw
from marge.autotuning import autotuning


class MarcosController(MarcosToolBar):
    """
    Controller class for managing MaRCoS functionality.
    """

    def __init__(self, *args, **kwargs):
        super(MarcosController, self).__init__(*args, **kwargs)

        # Copy relevant files from marcos_extras
        extras_path = os.path.join(os.path.dirname(__file__), "..", "marcos", "marcos_extras")
        dst = os.path.join(os.path.dirname(__file__), "..", "MaRGE")
        os.makedirs(dst, exist_ok=True)

        files_to_copy = [
            "copy_bitstream.sh",
            "marcos_fpga_rp-122.bit",
            "marcos_fpga_rp-122.bit.bin",
            "marcos_fpga_rp-122.dtbo",
            "readme.org"
        ]

        for fname in files_to_copy:
            src_file = os.path.join(extras_path, fname)
            if os.path.exists(src_file):
                shutil.copy(src_file, dst)
            else:
                print(f"[WARNING] File not found and not copied: {src_file}")

        self.action_server.setCheckable(True)
        self.action_start.triggered.connect(self.startMaRCoS)
        self.action_server.triggered.connect(self.controlMarcosServer)
        self.action_copybitstream.triggered.connect(self.copyBitStream)
        self.action_gpa_init.triggered.connect(self.initgpa)  # <- Tu as eu l'erreur ici

        thread = threading.Thread(target=self.search_sdrlab)
        thread.start()

        # Arduino to control the interlock
        self.arduino = autotuning.Arduino(baudrate=19200, name="interlock", serial_number=hw.ard_sn_interlock)
        self.arduino.connect()

    def search_sdrlab(self):
        if not self.main.demo:
            try:
                hw.rp_ip_address = self.get_sdrlab_ip()[0]
            except:
                print("ERROR: No SDRLab found.")
                try:
                    hw.rp_ip_address = self.get_sdrlab_ip()[0]
                except:
                    print("ERROR: No communication with SDRLab.")
                    print("ERROR: Try manually.")

    @staticmethod
    def get_sdrlab_ip():
        print("Searching for SDRLabs...")
        ip_addresses = []
        subnet = '192.168.1.'
        timeout = 0.1

        for i in range(101, 132):
            ip = subnet + str(i)
            try:
                if platform.system() == 'Linux':
                    result = subprocess.run(['ping', '-c', '1', ip], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, timeout=timeout)
                elif platform.system() == 'Windows':
                    result = subprocess.run(['ping', '-n', '1', ip], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, timeout=timeout)
                else:
                    continue

                if result.returncode == 0:
                    print(f"Checking ip {ip}...")
                    ssh_command = ['ssh', '-o', 'BatchMode=yes', '-o', 'ConnectTimeout=5', f'root@{ip}', 'exit']
                    ssh_result = subprocess.run(ssh_command, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
                    if ssh_result.returncode == 0:
                        ip_addresses.append(ip)
                    else:
                        print(f"WARNING: No SDRLab found at ip {ip}.")
            except:
                continue

        for ip in ip_addresses:
            print("READY: SDRLab found at IP " + ip)

        return ip_addresses

    def startMaRCoS(self):
        if not self.main.demo:
            try:
                subprocess.run([hw.bash_path, "--", "./communicateRP.sh", hw.rp_ip_address, "killall marcos_server"])
                subprocess.run([hw.bash_path, "--", "./startRP.sh", hw.rp_ip_address, hw.rp_version])
                self.initgpa()
                print("READY: MaRCoS updated, server connected, gpa initialized.")
            except:
                print("ERROR: Server connection not found! Please verify if the blue LED is illuminated on the Red Pitaya.")
        else:
            print("This is a demo\n")

        self.action_server.setChecked(True)
        self.main.toolbar_sequences.serverConnected()

    def controlMarcosServer(self):
        if not self.main.demo:
            if not self.action_server.isChecked():
                subprocess.run([hw.bash_path, "--", "./communicateRP.sh", hw.rp_ip_address, "killall marcos_server"])
                self.action_server.setStatusTip('Connect to marcos server')
                self.action_server.setToolTip('Connect to marcos server')
                print("Server disconnected")
            else:
                try:
                    subprocess.run([hw.bash_path, "--", "./communicateRP.sh", hw.rp_ip_address, "killall marcos_server"])
                    time.sleep(1.5)
                    subprocess.run([hw.bash_path, "--", "./communicateRP.sh", hw.rp_ip_address, "~/marcos_server"])
                    time.sleep(1.5)
                    self.action_server.setStatusTip('Kill marcos server')
                    self.action_server.setToolTip('Kill marcos server')

                    expt = ex.Experiment(init_gpa=False)
                    expt.add_flodict({'grad_vx': (np.array([100]), np.array([0]))})
                    expt.run()
                    expt.__del__()

                    print("READY: Server connected!")
                except Exception as e:
                    print("ERROR: Server not connected!")
                    print("ERROR: Try to connect to the server again.")
                    print(e)
        else:
            print("This is a demo\n")

    def copyBitStream(self):
        if not self.main.demo:
            try:
                subprocess.run([hw.bash_path, "--", "./communicateRP.sh", hw.rp_ip_address, "killall marcos_server"])
                subprocess.run([hw.bash_path, '--', './copy_bitstream.sh', hw.rp_ip_address, 'rp-122'], timeout=10)
                print("READY: MaRCoS updated")
            except subprocess.TimeoutExpired as e:
                print("ERROR: MaRCoS init timeout")
                print(e)
        else:
            print("This is a demo\n")

        self.action_server.setChecked(False)
        self.main.toolbar_sequences.serverConnected()

    def initgpa(self):
        """
        Initializes the GPA (Gradient Power Amplifier) hardware.
        """
        if not self.main.demo:
            try:
                subprocess.run([hw.bash_path, "--", "./initGPA.sh", hw.rp_ip_address], timeout=10)
                print("READY: GPA initialized")
            except subprocess.TimeoutExpired as e:
                print("ERROR: GPA init timeout")
                print(e)
        else:
            print("This is a demo\n")



#!/usr/bin/env python3
# Basic script to plot the CSV that marga_sim produces, so that you can visualise the expected pulse sequence from the hardware.

import numpy as np
import matplotlib.pyplot as plt
import sys, pdb
from local_config import fpga_clk_freq_MHz
st = pdb.set_trace

if __name__ == "__main__":
    if len(sys.argv) < 2:
        print("Usage: python {:s} <csv_file.csv>".format(sys.argv[0]) )
        print("\t Example: python {:s} /tmp/marga_sim.csv".format(sys.argv[0]) )
        exit()

    # data = np.genfromtxt(sys.argv[1],
    #                      dtype=None,
    #                      comments='#',
    #                      delimiter=',',
    #                      skip_header=0,
    #                      usecols=(0,1),
    #                      names=True)

    data = np.loadtxt(sys.argv[1], skiprows=2, delimiter=',')
    data[1:, 0] = data[1:, 0] - data[1, 0] + 1 # remove dead time in the beginning taken up by simulated memory writes

    time_us = data[:,0]/fpga_clk_freq_MHz
    tx = data[:,1:5].astype(np.int16) / 32768
    # offset binary
    fhdo = data[:,5:9].astype(np.uint16) / 32768 - 1
    # 2's complement
    ocra1 = ( (data[:,9:13].astype(np.int32) ^ (1 << 17)) - (1 << 17) ).astype(np.int32) / 131072
    gdata = np.hstack([fhdo, ocra1])
    gdata_nonzero = np.nonzero(gdata.any(0))[0]

    rx = data[:,14:21].astype(np.uint8)
    rx_en = rx[:, 5:] # ignore the rate logic, only plot the RX enables
    io = data[:,21:].astype(float)
    io[:,-1] = io[:,-1] / 256 # scale down LEDs to [0, 1)

    fig, axes = plt.subplots(4, 1, figsize=(12,8), sharex='col')

    (txs, grads, rxs, ios) = axes

    txs.step(time_us, tx, where='post')
    txs.legend(['tx0 i', 'tx0 q', 'tx1 i', 'tx1 q'])

    glegends = ['fhdo x', 'fhdo y', 'fhdo z', 'fhdo z2', 'ocra1 x', 'ocra1 y', 'ocra1 z', 'ocra1 z2']

    if gdata_nonzero.size != 0:
        grads.step(time_us, gdata[:, gdata_nonzero], where='post')
        grads.legend([glegends[k] for k in gdata_nonzero])

    rxs.step(time_us, rx_en, where='post')
    rxs.legend(["rx0 en", "rx1 en"])

    ios.step(time_us, io, where='post')
    ios.legend(['tx gate', 'rx gate', 'trig out', 'leds'])
    ios.set_xlabel('time (us)')

    for ax in axes:
        ax.grid(True)

    fig.tight_layout()
    plt.show()



"""
@author:    José Miguel Algarín
@email:     josalggui@i3m.upv.es
@affiliation:MRILab, i3M, CSIC, Valencia, Spain
"""
from PyQt5.QtWidgets import QTabWidget, QSizePolicy, QWidget, QVBoxLayout


class CustomAndProtocolWidget(QTabWidget):
    def __init__(self):
        super().__init__()
        self.setSizePolicy(QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum))
        self.setMaximumWidth(400)

        # Create tabs for custom sequences and protocol
        self.custom_widget = QWidget()
        self.custom_layout = QVBoxLayout()
        self.custom_widget.setLayout(self.custom_layout)
        self.protocol_widget = QWidget()
        self.protocol_layout = QVBoxLayout()
        self.protocol_widget.setLayout(self.protocol_layout)

        self.addTab(self.custom_widget, "Custom")
        self.addTab(self.protocol_widget, "Protocols")

